from typing import Dict, List, Union

from docarray import Document, DocumentArray
from docarray.score import NamedScore
from numpy import dot
from numpy.linalg import norm


class ESConverter:
    def __init__(self):
        pass

    def convert_doc_map_to_es(
        self,
        docs_map: Dict[str, DocumentArray],
        index_name: str,
        encoder_to_fields: dict,
    ) -> List[Dict]:
        """
        Transform a dictionary (mapping encoder to DocumentArray) into a list of Elasticsearch documents.
        The `docs_map` dictionary is expected to have the following structure:
        {
            'encoder1': DocumentArray([...]),
            'encoder2': DocumentArray([...]), # same number of documents as encoder1
            ...
        }

        :param docs_map: dictionary mapping encoder to DocumentArray.
        :param index_name: name of the index to be used in Elasticsearch.
        :param encoder_to_fields: dictionary mapping encoder to fields.
        :return: a list of Elasticsearch documents as dictionaries ready to be indexed.
        """
        es_docs = {}
        for executor_name, documents in docs_map.items():
            for doc in documents:
                if doc.id not in es_docs:
                    es_docs[doc.id] = self.get_base_es_doc(doc, index_name)
                    _doc = DocumentArray(Document(doc, copy=True))
                    # remove embeddings from serialized doc
                    _doc[..., 'embedding'] = None
                    _doc[..., 'tensor'] = None
                    _doc[..., 'blob'] = None
                    es_docs[doc.id]['serialized_doc'] = _doc[0].to_base64()
                es_doc = es_docs[doc.id]
                for encoded_field in encoder_to_fields[executor_name]:
                    field_doc = getattr(doc, encoded_field)
                    es_doc[
                        f'{encoded_field}-{executor_name}.embedding'
                    ] = field_doc.embedding
                    if hasattr(field_doc, 'text') and field_doc.text:
                        es_doc['bm25_text'] += field_doc.text + ' '
                        es_doc['text'] = field_doc.text
                    if hasattr(field_doc, 'uri') and field_doc.uri:
                        es_doc['uri'] = field_doc.uri
        return list(es_docs.values())

    def get_base_es_doc(self, doc: Document, index_name: str) -> Dict:
        es_doc = {k: v for k, v in doc.to_dict().items() if v}
        es_doc.pop('chunks', None)
        es_doc.pop('_metadata', None)
        es_doc['bm25_text'] = self.get_bm25_fields(doc)
        es_doc['_op_type'] = 'index'
        es_doc['_index'] = index_name
        es_doc['_id'] = doc.id
        doc.tags['embeddings'] = {}
        return es_doc

    def get_bm25_fields(self, doc: Document) -> str:
        try:
            return doc.bm25_text.text
        except AttributeError:
            return ''

    def convert_es_to_da(
        self, result: Union[Dict, List[Dict]], get_score_breakdown: bool
    ) -> DocumentArray:
        """
        Transform Elasticsearch documents into DocumentArray. Assumes that all Elasticsearch
        documents have a 'text' field. It returns embeddings as part of the tags for each field that is encoded.

        :param result: results from an Elasticsearch query.
        :param get_score_breakdown: whether to return the embeddings as tags for each document.
        :return: a DocumentArray containing all results.
        """
        if isinstance(result, Dict):
            result = [result]
        da = DocumentArray()
        for es_doc in result:
            doc = Document.from_base64(es_doc['_source']['serialized_doc'])
            for k, v in es_doc['_source'].items():
                if (
                    k.startswith('embedding') or k.endswith('embedding')
                ) and get_score_breakdown:
                    if 'embeddings' not in doc.tags:
                        doc.tags['embeddings'] = {}
                    doc.tags['embeddings'][k] = v
            da.append(doc)
        return da

    def convert_es_results_to_matches(
        self,
        query_doc: Document,
        es_results: List[Dict],
        get_score_breakdown: bool,
        metric: str,
        semantic_scores,
    ) -> DocumentArray:
        """
        Transform a list of results from Elasticsearch into a matches in the form of a `DocumentArray`.

        :param es_results: List of dictionaries containing results from Elasticsearch querying.
        :param get_score_breakdown: whether to calculate the score breakdown for matches.
        :return: `DocumentArray` that holds all matches in the form of `Document`s.
        """
        matches = DocumentArray()
        for result in es_results:
            d = self.convert_es_to_da(result, get_score_breakdown)[0]
            d.scores[metric] = NamedScore(value=result['_score'])
            if get_score_breakdown:
                d = self.calculate_score_breakdown(
                    query_doc, d, semantic_scores, metric
                )
            d.embedding = None
            d.tensor = None
            d.blob = b''
            matches.append(d)
        return matches

    def calculate_score_breakdown(
        self, query_doc: Document, retrieved_doc: Document, semantic_scores, metric
    ) -> Document:
        """
        Calculate the score breakdown for a given retrieved document. Each SemanticScore in the indexers
        `default_semantic_scores` should have a corresponding value, returned inside a list of scores in the documents
        tags under `score_breakdown`.

        :param query_doc: The query document. Contains embeddings for the semantic score calculation at tag level.
        :param retrieved_results: The Elasticsearch results, containing embeddings inside the `_source` field.
        :return: List of integers representing the score breakdown.
        """
        for (
            query_field,
            document_field,
            encoder,
            linear_weight,
        ) in semantic_scores:
            if encoder == 'bm25':
                continue
            q_emb = query_doc.tags['embeddings'][f'{query_field}-{encoder}']
            d_emb = retrieved_doc.tags['embeddings'][
                f'{document_field}-{encoder}.embedding'
            ]
            if metric == 'cosine':
                score = self.calculate_cosine(d_emb, q_emb) * linear_weight
            elif metric == 'l2_norm':
                score = self.calculate_l2_norm(d_emb, q_emb) * linear_weight
            else:
                raise ValueError(f'Invalid metric {metric}')
            retrieved_doc.scores[
                '-'.join(
                    [
                        query_field,
                        document_field,
                        encoder,
                        str(linear_weight),
                    ]
                )
            ] = NamedScore(value=round(score, 6))

        # calculate bm25 score
        vector_total = sum(
            [v.value for k, v in retrieved_doc.scores.items() if k != metric]
        )
        overall_score = retrieved_doc.scores[metric].value
        bm25_normalized = overall_score - vector_total
        bm25_raw = (bm25_normalized - 1) * 10

        retrieved_doc.scores['bm25_normalized'] = NamedScore(
            value=round(bm25_normalized, 6)
        )
        retrieved_doc.scores['bm25_raw'] = NamedScore(value=round(bm25_raw, 6))

        # remove embeddings from document
        retrieved_doc.tags.pop('embeddings', None)
        return retrieved_doc

    def calculate_l2_norm(self, d_emb, q_emb):
        return norm(q_emb - d_emb)

    def calculate_cosine(self, d_emb, q_emb):
        return dot(q_emb, d_emb) / (norm(q_emb) * norm(d_emb))



from typing import Optional

from pydantic import BaseModel, Field, root_validator


class ModalityModel(BaseModel):
    uri: Optional[str] = Field(default=None, description='URI of the file or data URI')
    text: Optional[str] = Field(default=None, description='Text')
    blob: Optional[str] = Field(
        default=None, description='Base64 encoded `utf-8` str format'
    )

    @root_validator(pre=True)
    def validate_only_one_exists(cls, values):
        # Get the names of all fields that are set (i.e. have a non-None value)
        set_fields = [name for name, value in values.items() if value is not None]
        if len(set_fields) != 1:
            raise ValueError(f"Only one of {set_fields} can be set.")
        return values



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field

from deployment.bff.app.v1.models.helper import BaseRequestModel
from deployment.bff.app.v1.models.modality import ModalityModel

_ProtoValueType = Optional[Union[bool, float, str, list, dict]]
_StructValueType = Union[
    _ProtoValueType, List[_ProtoValueType], Dict[str, _ProtoValueType]
]


class _NamedScore(BaseModel):
    value: Optional[float] = None


class IndexRequestModel(BaseRequestModel):
    data: List[Tuple[Dict[str, ModalityModel], Dict[str, Any]]] = Field(
        default=[({}, {})],
        description='List of tuples where each tuple contains a dictionary of data and a dictionary of tags. '
        'The data dictionary maps the field name to its value. ',
    )


class SearchRequestModel(BaseRequestModel):
    limit: int = Field(default=10, description='Number of matching results to return')
    filters: Optional[Dict[str, str]] = Field(
        default={},
        description='dictionary with filters for search results  {"tag_name" : "tag_value"}',
    )
    query: Dict[str, ModalityModel] = Field(
        default={}, description='Dictionary which maps the field name to its value. '
    )


class SearchResponseModel(BaseModel):
    id: str = Field(
        default=..., nullable=False, description='Id of the matching result.'
    )
    scores: Optional[Dict[str, '_NamedScore']] = Field(
        description='Similarity score with respect to the query.'
    )
    tags: Optional[Dict[str, '_StructValueType']] = Field(
        description='Additional tags associated with the file.'
    )
    fields: Dict[str, ModalityModel] = Field(
        default={}, description='Dictionary which maps the field name to its value. '
    )

    class Config:
        case_sensitive = False
        arbitrary_types_allowed = True


class SuggestionRequestModel(BaseRequestModel):
    text: Optional[str] = Field(default=None, description='Text')


IndexRequestModel.update_forward_refs()
SearchRequestModel.update_forward_refs()
SearchResponseModel.update_forward_refs()



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import importlib
import os
import uuid

from hubble import AuthenticationRequiredError
from kubernetes import client, config

from now.common.detect_schema import (
    set_field_names_from_docarray,
    set_field_names_from_local_folder,
    set_field_names_from_s3_bucket,
)
from now.constants import Apps, DatasetTypes
from now.demo_data import AVAILABLE_DATASETS
from now.deployment.deployment import cmd
from now.log import yaspin_extended
from now.now_dataclasses import DialogOptions, UserInput
from now.utils import (
    _get_context_names,
    get_info_hubble,
    jina_auth_login,
    sigmap,
    to_camel_case,
)

NEW_CLUSTER = {'name': '🐣 create new', 'value': 'new'}
AVAILABLE_SOON = 'will be available in upcoming versions'


# Make sure you add this dialog option to your app in order of dependency, i.e., if some dialog option depends on other
# than the parent should be called first before the dependant can called.


def _create_app_from_user_input(user_input: UserInput, **kwargs):
    if len(user_input.search_fields) != 1:
        raise ValueError(
            'Currently only one search field is supported. Please choose one field.'
        )
    if user_input.search_fields[0] not in user_input.search_fields_modalities.keys():
        raise ValueError(
            f'Search field specified is not among the search candidate fields. Please '
            f'choose one of the following: {user_input.search_fields_modalities.keys()}'
        )
    _search_modality = user_input.search_fields_modalities[user_input.search_fields[0]]
    if _search_modality in ['image', 'text', 'video']:
        app_name = Apps.SEARCH_APP
    else:
        raise ValueError(f'Invalid search modality: {_search_modality}')
    user_input.app_instance = construct_app(app_name)


APP_NAME = DialogOptions(
    name='flow_name',
    prompt_message='Choose a name for your application:',
    prompt_type='input',
    is_terminal_command=True,
    post_func=lambda user_input, **kwargs: clean_flow_name(user_input),
)


def clean_flow_name(user_input: UserInput):
    """
    Clean the flow name to make it valid, removing special characters and spaces.
    """
    user_input.flow_name = ''.join(
        [c for c in user_input.flow_name if c.isalnum() or c == '-']
    ).lower()


DATASET_TYPE = DialogOptions(
    name='dataset_type',
    prompt_message='How do you want to provide input? (format: https://docarray.jina.ai/)',
    choices=[
        {'name': 'Demo dataset', 'value': DatasetTypes.DEMO},
        {
            'name': 'DocumentArray name (recommended)',
            'value': DatasetTypes.DOCARRAY,
        },
        {
            'name': 'Local folder',
            'value': DatasetTypes.PATH,
        },
        {
            'name': 'S3 bucket',
            'value': DatasetTypes.S3_BUCKET,
        },
        {
            'name': 'Elasticsearch',
            'value': DatasetTypes.ELASTICSEARCH,
            'disabled': AVAILABLE_SOON,
        },
    ],
    prompt_type='list',
    is_terminal_command=True,
    post_func=lambda user_input, **kwargs: check_login_dataset(user_input),
)


def check_login_dataset(user_input: UserInput):
    if (
        user_input.dataset_type in [DatasetTypes.DEMO, DatasetTypes.DOCARRAY]
        and user_input.jwt is None
    ):
        _jina_auth_login(user_input)


def _get_demo_data_choices(user_input: UserInput, **kwargs):
    all_demo_datasets = []
    for demo_datasets in AVAILABLE_DATASETS.values():
        all_demo_datasets.extend(demo_datasets)
    return [
        {'name': demo_data.display_name, 'value': demo_data.name}
        for demo_data in all_demo_datasets
    ]


DEMO_DATA = DialogOptions(
    name='dataset_name',
    prompt_message='What demo dataset do you want to use?',
    choices=lambda user_input, **kwargs: _get_demo_data_choices(user_input),
    prompt_type='list',
    depends_on=DATASET_TYPE,
    is_terminal_command=True,
    description='Select one of the available demo datasets',
    conditional_check=lambda user_input, **kwargs: user_input.dataset_type
    == DatasetTypes.DEMO,
    post_func=lambda user_input, **kwargs: set_field_names_from_docarray(user_input),
)


DOCARRAY_NAME = DialogOptions(
    name='dataset_name',
    prompt_message='Please enter your DocumentArray name:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.DOCARRAY,
    post_func=lambda user_input, **kwargs: set_field_names_from_docarray(user_input),
)

DATASET_PATH = DialogOptions(
    name='dataset_path',
    prompt_message='Please enter the path to the local folder:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    is_terminal_command=True,
    conditional_check=lambda user_input: user_input.dataset_type == DatasetTypes.PATH,
    post_func=lambda user_input, **kwargs: set_field_names_from_local_folder(
        user_input
    ),
)

DATASET_PATH_S3 = DialogOptions(
    name='dataset_path',
    prompt_message='Please enter the S3 URI to the folder:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
)

AWS_ACCESS_KEY_ID = DialogOptions(
    name='aws_access_key_id',
    prompt_message='Please enter the AWS access key ID:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
)

AWS_SECRET_ACCESS_KEY = DialogOptions(
    name='aws_secret_access_key',
    prompt_message='Please enter the AWS secret access key:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
)

AWS_REGION_NAME = DialogOptions(
    name='aws_region_name',
    prompt_message='Please enter the AWS region:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.S3_BUCKET,
    post_func=lambda user_input, **kwargs: set_field_names_from_s3_bucket(user_input),
)

# --------------------------------------------- #


SEARCH_FIELDS = DialogOptions(
    name='search_fields',
    choices=lambda user_input, **kwargs: [
        {'name': field, 'value': field}
        for field in user_input.search_fields_modalities.keys()
    ],
    prompt_message='Please select the index fields:',
    prompt_type='checkbox',
    is_terminal_command=True,
    post_func=_create_app_from_user_input,
    argparse_kwargs={
        'type': lambda s: s.split(',') if s else UserInput().search_fields
    },
)


FILTER_FIELDS = DialogOptions(
    name='filter_fields',
    choices=lambda user_input, **kwargs: [
        {'name': field, 'value': field}
        for field in user_input.filter_fields_modalities.keys()
        if field not in user_input.search_fields
    ],
    prompt_message='Please select the filter fields',
    prompt_type='checkbox',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.filter_fields_modalities is not None
    and len(
        set(user_input.filter_fields_modalities.keys()) - set(user_input.search_fields)
    )
    > 0,
)


ES_INDEX_NAME = DialogOptions(
    name='es_index_name',
    prompt_message='Please enter the name of your Elasticsearch index:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.ELASTICSEARCH,
)


ES_HOST_NAME = DialogOptions(
    name='es_host_name',
    prompt_message='Please enter the address of your Elasticsearch node:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

ES_ADDITIONAL_ARGS = DialogOptions(
    name='es_additional_args',
    prompt_message='Please enter additional arguments for your Elasticsearch node if there are any:',
    prompt_type='input',
    depends_on=DATASET_TYPE,
    conditional_check=lambda user_input: user_input.dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

# --------------------------------------------- #

DEPLOYMENT_TYPE = DialogOptions(
    name='deployment_type',
    prompt_message='Where do you want to deploy your search engine?',
    prompt_type='list',
    choices=[
        {
            'name': '⛅️ Jina Cloud',
            'value': 'remote',
        },
        {
            'name': '📍 Local',
            'value': 'local',
        },
    ],
    is_terminal_command=True,
    description='Options are `local` or `remote`. Select `local` if you want your search engine to be deployed on a local '
    'cluster. Select `remote` to deploy it on Jina Cloud',
    post_func=lambda user_input, **kwargs: check_login_deployment(user_input),
)


def check_login_deployment(user_input: UserInput):
    if user_input.deployment_type == 'remote' and user_input.jwt is None:
        _jina_auth_login(user_input)


LOCAL_CLUSTER = DialogOptions(
    name='cluster',
    prompt_message='On which cluster do you want to deploy your search engine?',
    prompt_type='list',
    choices=lambda user_input, **kwargs: _construct_local_cluster_choices(
        user_input, **kwargs
    ),
    depends_on=DEPLOYMENT_TYPE,
    is_terminal_command=True,
    description='Reference an existing `local` cluster or select `new` to create a new one. '
    'Use this only when the `--deployment-type=local`',
    conditional_check=lambda user_inp: user_inp.deployment_type == 'local',
    post_func=lambda user_input, **kwargs: user_input.app_instance.run_checks(
        user_input
    ),
)


SECURED = DialogOptions(
    name='secured',
    prompt_message='Do you want to secure the Flow?',
    prompt_type='list',
    choices=[
        {'name': '⛔ no', 'value': False},
        {'name': '✅ yes', 'value': True},
    ],
    depends_on=DEPLOYMENT_TYPE,
    is_terminal_command=True,
    conditional_check=lambda user_inp: user_inp.deployment_type == 'remote',
)

API_KEY = DialogOptions(
    name='api_key',
    prompt_message='Do you want to generate an API key to access this deployment?',
    prompt_type='list',
    choices=[
        {'name': '✅ yes', 'value': uuid.uuid4().hex},
        {'name': '⛔ no', 'value': False},
    ],
    depends_on=SECURED,
    is_terminal_command=True,
    description='Pass an API key to access the Flow once the deployment is complete. ',
    conditional_check=lambda user_inp: str(user_inp.secured).lower() == 'true',
    post_func=lambda user_input, **kwargs: _set_value_to_none(user_input),
)

ADDITIONAL_USERS = DialogOptions(
    name='additional_user',
    prompt_message='Do you want to provide additional users access to this flow?',
    prompt_type='list',
    choices=[
        {'name': '✅ yes', 'value': True},
        {'name': '⛔ no', 'value': False},
    ],
    depends_on=SECURED,
    conditional_check=lambda user_inp: str(user_inp.secured).lower() == 'true',
)

USER_EMAILS = DialogOptions(
    name='user_emails',
    prompt_message='Please enter email addresses (separated by commas) '
    'to grant access to this Flow.\nAdditionally, you can specify comma-separated domain names'
    ' such that all users from that domain can access this Flow, e.g. `jina.ai`\n',
    prompt_type='input',
    depends_on=ADDITIONAL_USERS,
    conditional_check=lambda user_inp: user_inp.additional_user,
    post_func=lambda user_input, **kwargs: _add_additional_users(user_input, **kwargs),
)


def _set_value_to_none(user_input: UserInput):
    if not user_input.api_key:
        user_input.api_key = None


def _add_additional_users(user_input: UserInput, **kwargs):
    user_input.user_emails = (
        [email.strip() for email in kwargs['user_emails'].split(',')]
        if kwargs['user_emails']
        else []
    )


def _check_if_namespace_exist():
    config.load_kube_config()
    v1 = client.CoreV1Api()
    return 'nowapi' in [item.metadata.name for item in v1.list_namespace().items]


def construct_app(app_name: str):
    return getattr(
        importlib.import_module(f'now.app.{app_name}.app'),
        f'{to_camel_case(app_name)}',
    )()


def _jina_auth_login(user_input: UserInput, **kwargs):
    try:
        jina_auth_login()
    except AuthenticationRequiredError:
        with yaspin_extended(
            sigmap=sigmap, text='Log in to Jina AI Cloud', color='green'
        ) as spinner:
            cmd('jina auth login')
        spinner.ok('🛠️')

    get_info_hubble(user_input)
    os.environ['JCLOUD_NO_SURVEY'] = '1'


def _construct_local_cluster_choices(user_input: UserInput, **kwargs):
    active_context = kwargs.get('active_context')
    contexts = kwargs.get('contexts')
    context_names = _get_context_names(contexts, active_context)
    choices = [NEW_CLUSTER]
    # filter contexts with `gke`
    if len(context_names) > 0 and len(context_names[0]) > 0:
        context_names = [
            context
            for context in context_names
            if 'gke' not in context and _cluster_running(context)
        ]
        choices = context_names + choices
    return choices


def _cluster_running(cluster):
    config.load_kube_config(context=cluster)
    v1 = client.CoreV1Api()
    try:
        v1.list_namespace()
    except Exception as e:
        return False
    return True


app_config = [APP_NAME]
data_type = [DATASET_TYPE]
data_fields = [SEARCH_FIELDS, FILTER_FIELDS]
data_demo = [DEMO_DATA]
data_da = [DOCARRAY_NAME, DATASET_PATH]
data_s3 = [DATASET_PATH_S3, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_REGION_NAME]
data_es = [
    ES_HOST_NAME,
    ES_INDEX_NAME,
    ES_ADDITIONAL_ARGS,
]
cluster = [DEPLOYMENT_TYPE, LOCAL_CLUSTER]
remote_cluster = [SECURED, API_KEY, ADDITIONAL_USERS, USER_EMAILS]

base_options = (
    data_type
    + data_demo
    + data_da
    + data_s3
    + data_es
    + data_fields
    + app_config
    + cluster
    + remote_cluster
)



# -*- coding: utf-8 -*-
import json

__version__ = '0.1.2'


def format_json(data):
    return json.dumps(data, sort_keys=True, indent=4)


def colorize_json(data):
    try:
        from pygments import formatters, highlight, lexers

        if isinstance(data, bytes):
            data = data.decode('UTF-8')
        colorful_json = highlight(
            data, lexers.JsonLexer(), formatters.TerminalFormatter()
        )
        return colorful_json
    except ModuleNotFoundError:
        return data


def print_json(data):
    # colorful_json = highlight(unicode(format_json(data), 'UTF-8'),
    #                          lexers.JsonLexer(),
    #                          formatters.TerminalFormatter())
    print(colorize_json(format_json(data)))



from docarray import DocumentArray

from now.executor.abstract.base_indexer import NOWBaseIndexer


class InMemoryIndexer(NOWBaseIndexer):
    """InMemoryIndexer indexes Documents into a DocumentArray with `storage='memory'`"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print('InMemoryIndexer init')

    def construct(self, **kwargs):
        self._index = DocumentArray()

    def batch_iterator(self):
        for doc in self._index:
            yield doc

    def convert_filter_syntax(self, search_filter={}, search_filter_not={}):
        converted_filter = {}
        for key, val in search_filter.items():
            converted_filter[f"tags__{key}"] = val
        if search_filter_not:
            converted_filter['$not'] = {}
            for key_not, val_not in search_filter_not.items():
                converted_filter['$not'][f"tags__{key_not}"] = val_not
        return converted_filter

    def index(self, docs, parameters, **kwargs):
        for d in docs:
            if 'title' in d.tags:
                d.tags['title'] = d.tags['title'].lower()
        self._index.extend(docs)

    def delete(self, filtered_docs, *args, **kwargs):
        for doc in filtered_docs:
            del self._index[doc.id]

    def search(self, docs, parameters, retrieval_limit, search_filter, **kwargs):
        docs.match(self._index.find(search_filter), limit=retrieval_limit)



import os
from argparse import Namespace

from now.cli import cli
from now.constants import Apps, DatasetTypes

os.environ['JCLOUD_LOGLEVEL'] = 'DEBUG'


def deploy():
    kwargs = {
        'now': 'start',
        'app': Apps.SEARCH_APP,
        'dataset_type': DatasetTypes.S3_BUCKET,
        'dataset_path': '',
        'aws_access_key_id': '',
        'aws_secret_access_key': '',
        'aws_region_name': '',
        'deployment_type': 'remote',
        'proceed': True,
        'secured': True,
        'additional_user': True,
        'user_emails': [],
    }
    kwargs = Namespace(**kwargs)
    response = cli(args=kwargs)
    return response


if __name__ == '__main__':
    result = deploy()
    print(result)



import os
from typing import Dict, List, Tuple, Union

from docarray import DocumentArray

from now.app.base.app import JinaNOWApp
from now.common.utils import _get_clip_apps_with_dict, common_setup, get_indexer_config
from now.constants import CLIP_USES, Apps, Modalities
from now.demo_data import DemoDatasetNames
from now.now_dataclasses import UserInput


class SearchApp(JinaNOWApp):
    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.SEARCH_APP

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Image-text search app'

    @property
    def input_modality(self) -> Union[Modalities, List[Modalities]]:
        return [Modalities.TEXT, Modalities.IMAGE]

    @property
    def output_modality(self) -> Union[Modalities, List[Modalities]]:
        return [Modalities.TEXT, Modalities.IMAGE, Modalities.VIDEO]

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 8

    @property
    def finetune_datasets(self) -> [Tuple]:
        return (DemoDatasetNames.DEEP_FASHION, DemoDatasetNames.BIRD_SPECIES)

    def set_flow_yaml(self, **kwargs):
        now_package_dir = os.path.abspath(
            os.path.join(__file__, '..', '..', '..', '..')
        )
        flow_dir = os.path.join(now_package_dir, 'now', 'common', 'flow')
        self.flow_yaml = os.path.join(flow_dir, 'flow-clip.yml')

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        indexer_config = get_indexer_config()
        encoder_with = _get_clip_apps_with_dict(user_input)
        env_dict = common_setup(
            app_instance=self,
            user_input=user_input,
            dataset=dataset,
            encoder_uses=CLIP_USES[user_input.deployment_type][0],
            encoder_with=encoder_with,
            encoder_uses_with={
                'pretrained_model_name_or_path': CLIP_USES[user_input.deployment_type][
                    1
                ]
            },
            pre_trained_embedding_size=CLIP_USES[user_input.deployment_type][2],
            indexer_uses=indexer_config['indexer_uses'],
            kubectl_path=kubectl_path,
            indexer_resources=indexer_config['indexer_resources'],
        )
        super().setup(dataset=dataset, user_input=user_input, kubectl_path=kubectl_path)
        return env_dict



# TODO bff_request_mapping_fn and bff_response_mapping_fn should be used to create all routes

import logging.config
import sys

import uvicorn
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.routing import Mount

import deployment.bff.app.settings as api_settings
from deployment.bff.app.decorators import api_method, timed
from deployment.bff.app.v1.routers import admin, cloud_temp_link, search

logging.config.dictConfig(api_settings.DEFAULT_LOGGING_CONFIG)
logger = logging.getLogger('bff.app')
logger.setLevel(api_settings.DEFAULT_LOGGING_LEVEL)

TITLE = 'Jina NOW'
DESCRIPTION = 'The Jina NOW service API'
AUTHOR = 'Jina AI'
EMAIL = 'hello@jina.ai'
__version__ = 'latest'


def get_app_instance():
    """Build FastAPI app."""
    app = FastAPI(
        title=TITLE,
        description=DESCRIPTION,
        contact={
            'author': AUTHOR,
            'email': EMAIL,
        },
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )

    @app.get('/ping')
    @api_method
    @timed
    def check_liveness() -> str:
        """
        Sanity check - this will let the caller know that the service is operational.
        """
        return 'pong!'

    @app.get('/')
    @api_method
    @timed
    def read_root() -> str:
        """
        Root path welcome message.
        """
        return (
            f'{TITLE} v{__version__} 🚀 {DESCRIPTION} ✨ '
            f'author: {AUTHOR} email: {EMAIL} 📄  '
            'Check out /docs or /redoc for the API documentation!'
        )

    @app.on_event('startup')
    def startup():
        logger.info(
            f'Jina NOW started! ' f'Listening to [::]:{api_settings.DEFAULT_PORT}'
        )

    @app.exception_handler(Exception)
    async def unicorn_exception_handler(request: Request, exc: Exception):
        import traceback

        error = traceback.format_exc()
        return JSONResponse(
            status_code=500,
            content={
                "message": f"Exception in BFF, but the root cause can still be in the flow: {error}"
            },
        )

    return app


def build_app():
    # cloud temporary link router
    cloud_temp_link_mount = '/api/v1/cloud-bucket-utils'
    cloud_temp_link_app = get_app_instance()
    cloud_temp_link_app.include_router(
        cloud_temp_link.router, tags=['Temporary-Link-Cloud']
    )

    # search app router
    search_app_mount = '/api/v1/search-app'
    search_app_app = get_app_instance()
    search_app_app.include_router(search.router, tags=['Search App'])

    # Admin router
    admin_mount = '/api/v1/admin'
    admin_app = get_app_instance()
    admin_app.include_router(admin.router, tags=['admin'])

    # Mount them - for other modalities just add an app instance
    app = Starlette(
        routes=[
            Mount(cloud_temp_link_mount, cloud_temp_link_app),
            Mount(search_app_mount, search_app_app),
            Mount(admin_mount, admin_app),
        ]
    )
    return app


application = build_app()


def run_server(port=8080):
    """Run server."""
    app = build_app()
    uvicorn.run(
        app,
        host='0.0.0.0',
        port=port,
        loop='uvloop',
        http='httptools',
    )


if __name__ == '__main__':
    try:
        run_server(9090)
    except Exception as exc:
        logger.critical(str(exc))
        logger.exception(exc)
        sys.exit(1)



import itertools
import json
import os
from collections import defaultdict
from copy import deepcopy
from sys import maxsize
from typing import List, Optional

from docarray import Document, DocumentArray

from now.constants import (
    ACCESS_PATHS,
    TAG_INDEXER_DOC_HAS_TEXT,
    TAG_OCR_DETECTOR_TEXT_IN_DOC,
)
from now.executor.abstract.auth import (
    SecurityLevel,
    get_auth_executor_class,
    secure_request,
)
from now.executor.abstract.base_indexer.ranking import merge_matches_sum

CLOUD_BUCKET_PREFIXES = ['s3://']

Executor = get_auth_executor_class()


class NOWBaseIndexer(Executor):
    def __init__(
        self,
        dim: int = None,
        columns: Optional[List] = None,
        metric: str = 'cosine',
        limit: int = 10,
        max_values_per_tag: int = 10,
        *args,
        **kwargs,
    ):
        """
        :param dim: Dimensionality of vectors to index.
        :param columns: List of tuples of the form (column_name, str_type). Here str_type must be a string that can be
        parsed as a valid Python type.
        :param metric: Distance metric type. Can be 'euclidean', 'inner_product', or 'cosine'
        :param limit: Number of results to get for each query document in search
        :param max_values_per_tag: Maximum number of values per tag
        (used for search), e.g. '@r', '@c', '@r,c'
        """

        super().__init__(*args, **kwargs)
        self.columns = self.parse_columns(columns)
        self.dim = dim
        self.metric = metric
        self.limit = limit
        self.max_values_per_tag = max_values_per_tag
        self.construct(**kwargs)
        self.doc_id_tags = {}
        self.document_list = DocumentArray()
        self.load_document_list()
        self.query_to_curated_matches_path = (
            os.path.join(self.workspace, 'query_to_curated_matches.json')
            if self.workspace
            else None
        )
        self.query_to_curated_matches = self.open_query_to_curated_matches(
            self.query_to_curated_matches_path
        )

    def open_query_to_curated_matches(self, path):
        if path and os.path.exists(path):
            with open(path, 'r') as f:
                return json.load(f)
        return defaultdict(list)

    @secure_request(on='/tags', level=SecurityLevel.USER)
    def get_tags_and_values(self, **kwargs):
        """Returns tags and their possible values.

        for example if indexed docs are the following:
            docs = DocumentArray([
                Document(.., tags={'color':'red'}),
                Document(.., tags={'color':'blue'}),
                Document(.., tags={'greeting':'hello'}),
            ])

        the resulting response would be a document array with
        one document containing a dictionary in tags like the following:
        {'tags':{'color':['red', 'blue'], 'greeting':['hello']}}.
        """
        count_dict = defaultdict(lambda: defaultdict(int))
        for tags in self.doc_id_tags.values():
            for key, value in tags.items():
                count_dict[key][value] += 1
        tag_to_values = dict()
        for key, value_to_count in count_dict.items():
            sorted_values = sorted(
                value_to_count.items(), reverse=True, key=lambda item: item[1]
            )
            tag_to_values[key] = [
                value
                for (value, _), _ in zip(sorted_values, range(self.max_values_per_tag))
            ]
        return DocumentArray([Document(text='tags', tags={'tags': tag_to_values})])

    @secure_request(on='/list', level=SecurityLevel.USER)
    def list(self, parameters: dict = {}, **kwargs):
        """List all indexed documents
        :param parameters: dictionary with limit and offset
        - offset (int): number of documents to skip
        - limit (int): number of retrieved documents
        """
        limit = int(parameters.get('limit', maxsize))
        offset = int(parameters.get('offset', 0))
        # add removal of duplicates
        parent_ids = sorted({doc.parent_id for doc in self.document_list})
        # filter by offset and limit
        filtered_parent_ids = parent_ids[offset : offset + limit]
        # get the documents of filtered parent ids
        docs = DocumentArray(
            [doc for doc in self.document_list if doc.parent_id in filtered_parent_ids]
        )
        return docs

    @staticmethod
    def has_text(doc):
        # TODO: pop is an unwanted side effect cleaning up the docs should be done somewhere else
        text_in_doc = doc.tags.pop(TAG_OCR_DETECTOR_TEXT_IN_DOC, '')
        text_in_doc = text_in_doc.split(' ')
        text_in_doc = filter(lambda s: len(s) > 1 and s.isalnum(), text_in_doc)
        return len(list(text_in_doc)) > 0

    @staticmethod
    def set_doc_has_text_tag(docs: DocumentArray):
        """
        Mark documents with the tag TAG_INDEXER_DOC_HAS_TEXT.
        If two documents have the same parent, then they are all marked as having text.
        """
        parent_ids = {d.parent_id for d in docs if NOWBaseIndexer.has_text(d)}
        for doc in docs:
            doc.tags[TAG_INDEXER_DOC_HAS_TEXT] = doc.parent_id in parent_ids

    @secure_request(on='/index', level=SecurityLevel.USER)
    def index_endpoint(
        self,
        docs: Optional[DocumentArray] = None,
        parameters: dict = {},
        docs_matrix: Optional[List[DocumentArray]] = None,
        **kwargs,
    ):
        """Base function for indexing documents. Handles the data management for the index and list endpoints.

        :param docs: the Documents to index
        :param parameters: dictionary with options for indexing
        """
        flat_docs = docs[ACCESS_PATHS]
        # TODO please remove this check for empty docs
        if len(flat_docs) == 0:
            return
        flat_docs = DocumentArray(
            [document for document in flat_docs if document.embedding is not None]
        )
        self.set_doc_has_text_tag(flat_docs)
        flat_docs = self.maybe_drop_blob_tensor(flat_docs)
        self.index(flat_docs, parameters, **kwargs)
        self.extend_inmemory_docs_and_tags(flat_docs)
        return DocumentArray([])

    @secure_request(on='/delete', level=SecurityLevel.USER)
    def delete_endpoint(self, parameters: dict = {}, **kwargs):
        """
        Delete endpoint to delete document/documents from the index.
        Filter conditions can be passed to select documents for deletion.
        """
        filter = parameters.get("filter", {})
        if filter:
            filtered_docs = deepcopy(self.document_list.find(filter=filter))
            self.delete_inmemory_docs_and_tags(filtered_docs)
            self.delete(filtered_docs, parameters, **kwargs)

        return DocumentArray()

    @secure_request(on='/search', level=SecurityLevel.USER)
    def search_endpoint(
        self,
        docs: Optional[DocumentArray] = None,
        parameters: dict = {},
        **kwargs,
    ):
        """Perform a vector similarity search and retrieve `Document` matches."""
        limit = int(parameters.get('limit', self.limit))
        search_filter_raw = parameters.get('filter', {})
        search_filter_orig = deepcopy(search_filter_raw)
        docs = docs[ACCESS_PATHS][:1]  # only search on the first document for now
        # TODO remove this check for empty docs and make sure everything else works
        if len(docs) == 0:
            return
        retrieval_limit = limit * 3

        # if OCR detector was used to check if documents contain text in indexed image modality adjust retrieval step
        if self.columns and TAG_INDEXER_DOC_HAS_TEXT in [
            col[0] for col in self.columns
        ]:
            # enhance filter for one word queries to include the title
            if (
                len(docs[0].text.split()) == 1
                and not search_filter_raw
                and 'title' in [col[0] for col in self.columns]
            ):
                search_filter_raw = {'title': {'$regex': docs[0].text.lower()}}
            # search for documents which don't contain text
            search_filter_raw[TAG_INDEXER_DOC_HAS_TEXT] = {'$eq': False}
            search_filter = self.convert_filter_syntax(search_filter_raw)
            docs_with_matches_no_text = self.create_matches(
                docs,
                parameters,
                limit,
                retrieval_limit,
                search_filter,
            )
            # search for documents which contain text
            search_filter_raw[TAG_INDEXER_DOC_HAS_TEXT] = {'$eq': True}
            search_filter = self.convert_filter_syntax(search_filter_raw)
            docs_with_matches_filter_title = self.create_matches(
                docs,
                parameters,
                limit,
                retrieval_limit,
                search_filter,
            )
            # merge the results such that documents with text are retrieved at a later stage
            self.merge_matches_by_score_after_half(
                docs_with_matches_no_text,
                docs_with_matches_filter_title,
                limit,
            )
            docs_with_matches = docs_with_matches_no_text
            # if we didn't retrieve enough results, try to fetch more
            if len(docs_with_matches) < limit:
                search_filter = self.convert_filter_syntax(search_filter_orig)
                docs_with_additonal_matches = self.create_matches(
                    docs,
                    parameters,
                    limit,
                    retrieval_limit,
                    search_filter,
                )
                self.append_matches_if_not_exists(
                    docs_with_matches, docs_with_additonal_matches, limit
                )
        else:
            search_filter = self.convert_filter_syntax(search_filter_orig)
            docs_with_matches = self.create_matches(
                docs, parameters, limit, retrieval_limit, search_filter
            )
        docs_with_matches[0].matches = (
            self.get_curated_matches(docs[0].text) + docs_with_matches[0].matches
        )
        self.remove_duplicates(
            docs_with_matches
        )  # TODO combine with append_matches_if_not_exists - duplicate code
        docs_with_matches[0].matches = docs_with_matches[0].matches[:limit]
        self.clean_response(docs_with_matches)
        return docs_with_matches

    def remove_duplicates(self, docs_with_matches):
        """Remove duplicate matches from the list of documents."""
        parent_ids = set()
        # curated matches can lead to duplicates since the document_list is on frame/sentence level
        # TODO simplify this logic by letting the document_list be on root level
        unique_curated_matches = DocumentArray([])
        for match in docs_with_matches[0].matches:
            if match.parent_id not in parent_ids:
                parent_ids.add(match.parent_id)
                unique_curated_matches.append(match)
        docs_with_matches[0].matches = unique_curated_matches

    def merge_matches_by_score_after_half(
        self,
        docs_with_matches: DocumentArray,
        docs_with_matches_to_add: DocumentArray,
        limit: int,
    ):
        # get all parent_ids of the matches of the docs_with_matches
        parent_ids = set()
        for doc, doc_to_add in zip(docs_with_matches, docs_with_matches_to_add):
            score_name = (
                list(doc.matches[0].scores.keys())[0]
                if len(doc.matches) > 0
                else self.metric
            )
            for match in doc.matches[: limit // 2]:
                parent_ids.add(match.parent_id)

            possible_matches = deepcopy(doc.matches[limit // 2 :]) + doc_to_add.matches
            doc.matches = doc.matches[: limit // 2]
            ids_scores = [
                (match.id, match.scores[score_name].value) for match in possible_matches
            ]
            ids_scores.sort(key=lambda x: x[1], reverse='similarity' in score_name)
            for id, _ in ids_scores:
                _match = possible_matches[id]
                if _match.id not in parent_ids:
                    if len(doc.matches) >= limit:
                        break
                    doc.matches.append(_match)

    @secure_request(on='/curate', level=SecurityLevel.USER)
    def curate(self, parameters: dict = {}, **kwargs):
        """
        This endpoint is only relevant for text queries.
        It defines the top results for each query.
        `parameters` should have the following format:
        {
            'query_to_filter': {
                'query1': [
                    {'uri': {'$eq': 'uri1'}},
                    {'tags__internal_id': {'$eq': 'id1'}},
                ],
                'query2': [
                    {'uri': {'$eq': 'uri2'}},
                    {'tags__color': {'$eq': 'red'}},
                ],
            }
        }
        """

        query_to_filter = parameters['query_to_filter']
        for query, doc_filters in query_to_filter.items():
            # a query can have multiple filters
            # for each filter, we get the curated matches
            # all matches are flattened into one list
            curated_matches = DocumentArray(
                itertools.chain(
                    *[self.document_list.find(doc_filter) for doc_filter in doc_filters]
                )
            )
            self.query_to_curated_matches[query] = curated_matches.to_list()

        with open(self.query_to_curated_matches_path, 'w') as f:
            json.dump(self.query_to_curated_matches, f)

    def get_curated_matches(self, text_query: str = None) -> DocumentArray:
        """
        Get curated matches for a given text query.
        """
        curated_matches = DocumentArray([])
        if text_query:
            matches = self.query_to_curated_matches.get(text_query, [])
            curated_matches.extend(DocumentArray.from_list(matches))
        return curated_matches

    def create_matches(self, docs, parameters, limit, retrieval_limit, search_filter):
        docs_copy = deepcopy(docs)
        self.search(docs_copy, parameters, retrieval_limit, search_filter)
        merge_matches_sum(docs_copy, limit)
        return docs_copy

    def append_matches_if_not_exists(
        self, docs_with_matches, docs_with_matches_to_add, limit
    ):
        # get all parent_ids of the matches of the docs_with_matches
        parent_ids = set()
        for doc, doc_to_add in zip(docs_with_matches, docs_with_matches_to_add):
            for match in doc.matches:
                parent_ids.add(match.parent_id)

            # append matches to docs_with_matches if they are not already in the matches
            for match in doc_to_add.matches:
                if match.parent_id not in parent_ids:
                    if len(doc.matches) >= limit:
                        break
                    doc.matches.append(match)

    @staticmethod
    def maybe_drop_blob_tensor(docs: DocumentArray):
        """Drops `blob` or `tensor` from documents which have `uri` attribute set and
        whose 'uri' is a data-uri or is either in the cloud(S3) or can be loaded."""
        for doc in docs:
            if doc.uri:
                if doc.text:
                    continue
                else:
                    try:
                        if not doc.uri.startswith(f'data:{doc.mime_type}') and not any(
                            [
                                doc.uri.startswith(cloud_bucket_prefix)
                                for cloud_bucket_prefix in CLOUD_BUCKET_PREFIXES
                            ]
                        ):
                            doc.load_uri_to_blob(timeout=10)
                        doc.blob = None
                        doc.tensor = None
                        doc.mime_type = None
                    except Exception as e:  # noqa E722
                        continue
        return docs

    @staticmethod
    def clean_response(docs):
        """Removes embedding & OCR tags from the root level and also from the matches."""

        def _clean_response(doc: Document):
            """Inplace removes embedding & tags associated with OCR detection."""
            doc.embedding = None
            for _tag in [TAG_INDEXER_DOC_HAS_TEXT, TAG_OCR_DETECTOR_TEXT_IN_DOC]:
                doc.tags.pop(_tag, None)

        for doc in docs:
            _clean_response(doc)
            for match in doc.matches:
                _clean_response(match)

    @staticmethod
    def parse_columns(columns):
        """Parse the columns to index"""
        valid_input_columns = ['str', 'float', 'int', 'bool']
        if columns:
            corrected_list = []
            for i in range(0, len(columns), 2):
                # This conversion is needed for MMDocs
                if columns[i + 1] in ['text', 'stringValue']:
                    columns[i + 1] = 'str'
                elif columns[i + 1] in ['numberValue']:
                    columns[i + 1] = 'float'
                elif columns[i + 1] in ['boolValue']:
                    columns[i + 1] = 'bool'
                corrected_list.append((columns[i], columns[i + 1]))
            columns = corrected_list
            for n, t in columns:
                assert (
                    t.lower() in valid_input_columns
                ), f'column of type={t} is not supported. Supported types are {valid_input_columns}'
        return columns

    def load_document_list(self):
        """is needed for the list endpoint"""
        document_list = DocumentArray()
        for batch in self.batch_iterator():
            self.extend_inmemory_docs_and_tags(batch)
        self.document_list = DocumentArray(
            sorted([d for d in document_list], key=lambda x: x.id)
        )

    def extend_inmemory_docs_and_tags(self, batch):
        """Extend the in-memory DocumentArray with new documents"""
        for d in batch:
            tags = deepcopy(d.tags)
            for _tag in [TAG_INDEXER_DOC_HAS_TEXT, TAG_OCR_DETECTOR_TEXT_IN_DOC]:
                tags.pop(_tag, None)
            self.document_list.append(
                Document(id=d.id, uri=d.uri, tags=tags, parent_id=d.parent_id)
            )
            self.doc_id_tags[d.id] = tags

    def delete_inmemory_docs_and_tags(self, docs):
        """Delete documents from the in-memory DocumentArray"""
        for d in docs:
            del self.document_list[d.id]
            self.doc_id_tags.pop(d.id)

    def update_inmemory_docs_and_tags(self, docs):
        """Update documents in the in-memory DocumentArray"""
        self.delete_inmemory_docs_and_tags(docs)
        self.extend_inmemory_docs_and_tags(docs)

    def construct(self, **kwargs):
        """Calls the constructor of the specialized indexer"""
        raise NotImplementedError

    def batch_iterator(self):
        """Needs to be implemented in derived classes. Iterates over all documents in batches and yields them"""
        raise NotImplementedError

    def convert_filter_syntax(self, search_filter={}, search_filter_not={}):
        """Needs to be implemented in derived classes. Converts the filter syntax to the syntax of the specific indexer"""
        raise NotImplementedError

    def index(self, docs: DocumentArray, parameters: dict, **kwargs):
        """Needs to be implemented in derived classes. Indexes the documents"""
        raise NotImplementedError

    def delete(self, docs_to_delete, parameters: dict, **kwargs):
        """Needs to be implemented in derived classes. Deletes the documents"""
        raise NotImplementedError

    def search(
        self,
        docs: DocumentArray,
        parameters: dict,
        limit: int,
        search_filter: dict,
        **kwargs,
    ):
        """Needs to be implemented in derived classes. Searches the documents"""
        raise NotImplementedError



# TODO this should not be needed in the future
name_to_id_map = {
    'NOWPreprocessor': 'vlxgoxlv',
    'NOWQdrantIndexer16': '11yxatzh',
    'NOWElasticIndexer': 'mdtuyaaj',
    'NOWAutoCompleteExecutor2': 'w5w084h7',
}

id_to_name_map = {
    'vlxgoxlv': 'NOWPreprocessor',
    '11yxatzh': 'NOWQdrantIndexer16',
    'mdtuyaaj': 'NOWElasticIndexer',
    'w5w084h7': 'NOWAutoCompleteExecutor2',
}



import itertools
import json
import os
from typing import Dict, List

import requests

from now.constants import (
    AVAILABLE_MODALITIES_FOR_FILTER,
    AVAILABLE_MODALITIES_FOR_SEARCH,
    NOT_AVAILABLE_MODALITIES_FOR_FILTER,
    SUPPORTED_FILE_TYPES,
    Modalities,
)
from now.data_loading.data_loading import get_s3_bucket_and_folder_prefix
from now.now_dataclasses import UserInput


def _create_candidate_search_filter_fields(field_name_to_value):
    """
    Creates candidate search fields from the field_names for s3
    and local file path.
    A candidate search field is a field that we can detect its modality
    A candidate filter field is a field that we can't detect its modality,
    or it's modality is different from image, video or audio.

    :param field_name_to_value: dictionary
    """
    search_fields_modalities = {}
    filter_field_modalities = {}
    not_available_file_types_for_filter = list(
        itertools.chain(
            *[
                SUPPORTED_FILE_TYPES[modality]
                for modality in NOT_AVAILABLE_MODALITIES_FOR_FILTER
            ]
        )
    )
    for field_name, field_value in field_name_to_value.items():
        # we determine search modality
        for modality in AVAILABLE_MODALITIES_FOR_SEARCH:
            file_types = SUPPORTED_FILE_TYPES[modality]
            if field_name.split('.')[-1] in file_types:
                search_fields_modalities[field_name] = modality
                break
            elif field_name == 'uri' and field_value.split('.')[-1] in file_types:
                search_fields_modalities[field_name] = modality
                break
            elif field_name == 'text' and field_value:
                search_fields_modalities[field_name] = Modalities.TEXT
                break
        # we determine if it's a filter field
        if (
            field_name == 'uri'
            and field_value.split('.')[-1] not in not_available_file_types_for_filter
        ) or field_name.split('.')[-1] not in not_available_file_types_for_filter:
            filter_field_modalities[field_name] = str(field_value.__class__.__name__)

    if len(search_fields_modalities.keys()) == 0:
        raise ValueError(
            'No searchable fields found, please check documentation https://now.jina.ai'
        )
    return search_fields_modalities, filter_field_modalities


def _extract_field_candidates_docarray(response):
    """
    Downloads the first document in the document array and extracts field names from it
    if tags also exists then it extracts the keys from tags and adds to the field_names
    """
    search_modalities = {}
    filter_modalities = {}
    doc = requests.get(response.json()['data']['download']).json()[0]
    if (
        not doc.get('_metadata', None)
        or 'multi_modal_schema' not in doc['_metadata']['fields']
    ):
        raise RuntimeError(
            'Multi-modal schema is not provided. Please prepare your data following this guide - '
            'https://docarray.jina.ai/datatypes/multimodal/'
        )
    mm_schema = doc['_metadata']['fields']['multi_modal_schema']
    mm_fields = mm_schema['structValue']['fields']
    for field_name, value in mm_fields.items():
        if 'position' not in value['structValue']['fields']:
            raise ValueError(
                f'No modality found for the dataclass field: `{field_name}`. Please follow the steps in the '
                f'documentation to add modalities to your documents https://docarray.jina.ai/datatypes/multimodal/'
            )
        field_pos = value['structValue']['fields']['position']['numberValue']
        if not doc['chunks'][field_pos]['modality']:
            raise ValueError(
                f'No modality found for {field_name}. Please follow the steps in the documentation'
                f' to add modalities to your documents https://docarray.jina.ai/datatypes/multimodal/'
            )
        modality = doc['chunks'][field_pos]['modality'].lower()
        if modality not in AVAILABLE_MODALITIES_FOR_SEARCH:
            raise ValueError(
                f'The modality {modality} is not supported for search. Please use '
                f'one of the following modalities: {AVAILABLE_MODALITIES_FOR_SEARCH}'
            )
        # only the available modalities for filter are added to the filter modalities
        if modality in AVAILABLE_MODALITIES_FOR_FILTER:
            filter_modalities[field_name] = modality
        # only the available modalities for search are added to search modalities
        if modality in AVAILABLE_MODALITIES_FOR_SEARCH:
            search_modalities[field_name] = modality

    if doc.get('tags', None):
        for el, value in doc['tags']['fields'].items():
            for val_type, val in value.items():
                filter_modalities[el] = val_type

    if len(search_modalities.keys()) == 0:
        raise ValueError(
            'No searchable fields found, please check documentation https://now.jina.ai'
        )
    return search_modalities, filter_modalities


def set_field_names_from_docarray(user_input: UserInput, **kwargs):
    """
    Get the schema from a DocArray

    :param user_input: UserInput object

    Makes a request to hubble API and downloads the first 10 documents
    from the document array and uses the first document to get the schema and sets field_names in user_input
    """
    cookies = {
        'st': user_input.jwt['token'],
    }

    json_data = {
        'name': user_input.dataset_name,
    }
    response = requests.post(
        'https://api.hubble.jina.ai/v2/rpc/docarray.getFirstDocuments',
        cookies=cookies,
        json=json_data,
    )
    if response.json()['code'] == 200:
        (
            user_input.search_fields_modalities,
            user_input.filter_fields_modalities,
        ) = _extract_field_candidates_docarray(response)
    else:
        raise ValueError('DocumentArray does not exist or you do not have access to it')


def _identify_folder_structure(file_paths: List[str], separator: str) -> str:
    """This function identifies the folder structure.
    It works with a local file structure or a remote file structure.

    :param file_paths: list of file paths
    :param separator: separator used in the file paths
    :return: if all the files are in the same folder then returns 'single_folder' else returns 'sub_folders'
    :raises ValueError: if the files don't have the same depth in the file structure
    """
    # check if all files are in the same folder
    depths = [len(path.split(separator)) for path in file_paths]
    if len(set(depths)) != 1:
        raise ValueError(
            "Files have differing depth, please check documentation https://now.jina.ai"
        )
    # check if all files are in the same folder
    if (
        len(set([separator.join(path.split(separator)[:-1]) for path in file_paths]))
        != 1
    ):
        return 'sub_folders'
    return 'single_folder'


def _extract_field_names_single_folder(
    file_paths: List[str], separator: str
) -> Dict[str, str]:
    """This function extracts the file endings in a single folder and returns them as field names.
    It works with a local file structure or a remote file structure.

    :param file_paths: list of relative file paths from data set path
    :param separator: separator used in the file paths
    :return: list of file endings
    """
    file_endings = set(
        ['.' + path.split(separator)[-1].split('.')[-1] for path in file_paths]
    )
    return {file_ending: file_ending for file_ending in file_endings}


def _extract_field_names_sub_folders(
    file_paths: List[str], separator: str, s3_bucket=None
) -> Dict[str, str]:
    """This function extracts the files in sub folders and returns them as field names. Also, it reads json files
    and adds them as key-value pairs to the field names dictionary.
    It works with a local file structure or a remote file structure.

    :param file_paths: list of relative file paths from data set path
    :param separator: separator used in the file paths
    :param s3_bucket: s3 bucket object, only needed if interacting with s3 bucket
    :return: list of file endings
    """
    field_names = {}
    for path in file_paths:
        if path.endswith('.json'):
            if s3_bucket:
                data = json.loads(s3_bucket.Object(path).get()['Body'].read())
            else:
                with open(path) as f:
                    data = json.load(f)
            for el, value in data.items():
                field_names[el] = value
        else:
            file_name = path.split(separator)[-1]
            field_names[file_name] = file_name
    return field_names


def set_field_names_from_s3_bucket(user_input: UserInput, **kwargs):
    """
    Get the schema from a S3 bucket

    :param user_input: UserInput object

    checks if the bucket exists and the format of the folder structure is correct,
    if yes then downloads the first folder and sets its content as field_names in user_input
    """
    bucket, folder_prefix = get_s3_bucket_and_folder_prefix(
        user_input
    )  # user has to provide the folder where folder structure begins

    objects = list(bucket.objects.filter(Prefix=folder_prefix))
    file_paths = [obj.key for obj in objects if not obj.key.endswith('/')]
    folder_structure = _identify_folder_structure(file_paths, '/')
    if folder_structure == 'single_folder':
        field_names = _extract_field_names_single_folder(file_paths, '/')
    elif folder_structure == 'sub_folders':
        first_folder = '/'.join(objects[1].key.split('/')[:-1])
        first_folder_objects = [
            obj.key
            for obj in bucket.objects.filter(Prefix=first_folder)
            if not obj.key.endswith('/')
        ]
        field_names = _extract_field_names_sub_folders(
            first_folder_objects, '/', bucket
        )
    (
        user_input.search_fields_modalities,
        user_input.filter_fields_modalities,
    ) = _create_candidate_search_filter_fields(field_names)


def set_field_names_from_local_folder(user_input: UserInput, **kwargs):
    """
    Get the schema from a local folder

    :param user_input: UserInput object

    checks if the folder exists and the format of the folder structure is correct,
    if yes set the content of the first folder as field_names in user_input
    """
    dataset_path = user_input.dataset_path.strip()
    dataset_path = os.path.expanduser(dataset_path)
    if os.path.isfile(dataset_path):
        raise ValueError(
            'The path provided is not a folder, please check documentation https://now.jina.ai'
        )
    file_paths = []
    for root, dirs, files in os.walk(dataset_path):
        file_paths.extend([os.path.join(root, file) for file in files])
    folder_structure = _identify_folder_structure(file_paths, os.sep)
    if folder_structure == 'single_folder':
        field_names = _extract_field_names_single_folder(file_paths, os.sep)
    elif folder_structure == 'sub_folders':
        first_folder = os.sep.join(file_paths[0].split(os.sep)[:-1])
        first_folder_files = [
            os.path.join(first_folder, file)
            for file in os.listdir(first_folder)
            if os.path.isfile(os.path.join(first_folder, file))
        ]
        field_names = _extract_field_names_sub_folders(first_folder_files, os.sep)
    (
        user_input.search_fields_modalities,
        user_input.filter_fields_modalities,
    ) = _create_candidate_search_filter_fields(field_names)



from __future__ import annotations, print_function, unicode_literals

from docarray import DocumentArray
from pydantic import BaseModel

from now.constants import Modalities
from now.utils import BetterEnum


class DemoDatasetNames(BetterEnum):
    BEST_ARTWORKS = 'best-artworks'
    NFT_MONKEY = 'nft-monkey'
    TLL = 'totally-looks-like'
    BIRD_SPECIES = 'bird-species'
    STANFORD_CARS = 'stanford-cars'
    DEEP_FASHION = 'deepfashion'
    NIH_CHEST_XRAYS = 'nih-chest-xrays'
    GEOLOCATION_GEOGUESSR = 'geolocation-geoguessr'
    ROCK_LYRICS = 'rock-lyrics'
    POP_LYRICS = 'pop-lyrics'
    RAP_LYRICS = 'rap-lyrics'
    INDIE_LYRICS = 'indie-lyrics'
    METAL_LYRICS = 'metal-lyrics'
    TUMBLR_GIFS = 'tumblr-gifs'
    TUMBLR_GIFS_10K = 'tumblr-gifs-10k'
    ES_ONLINE_SHOP_50 = 'extracted-data-online-shop-50-flat'


class DemoDataset(BaseModel):
    name: str
    display_name: str
    display_modality: str

    def get_data(self, *args, **kwargs) -> DocumentArray:
        return DocumentArray.pull(self.name)


AVAILABLE_DATASETS = {
    Modalities.IMAGE: [
        DemoDataset(
            name=DemoDatasetNames.BEST_ARTWORKS,
            display_modality=Modalities.IMAGE,
            display_name='🖼 artworks (≈8K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.NFT_MONKEY,
            display_modality=Modalities.IMAGE,
            display_name='💰 nft - bored apes (10K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.TLL,
            display_modality=Modalities.IMAGE,
            display_name='👬 totally looks like (≈12K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.BIRD_SPECIES,
            display_modality=Modalities.IMAGE,
            display_name='🦆 birds (≈12K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.STANFORD_CARS,
            display_modality=Modalities.IMAGE,
            display_name='🚗 cars (≈16K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.GEOLOCATION_GEOGUESSR,
            display_modality=Modalities.IMAGE,
            display_name='🌍 geolocation (≈50K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.DEEP_FASHION,
            display_modality=Modalities.IMAGE,
            display_name='👕 fashion (≈53K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.NIH_CHEST_XRAYS,
            display_modality=Modalities.IMAGE,
            display_name='☢ chest x-rays (≈100K docs)',
        ),
    ],
    Modalities.TEXT: [
        DemoDataset(
            name=DemoDatasetNames.ROCK_LYRICS,
            display_modality=Modalities.TEXT,
            display_name='🎤 rock lyrics (200K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.POP_LYRICS,
            display_modality=Modalities.TEXT,
            display_name='🎤 pop lyrics (200K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.RAP_LYRICS,
            display_modality=Modalities.TEXT,
            display_name='🎤 rap lyrics (200K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.INDIE_LYRICS,
            display_modality=Modalities.TEXT,
            display_name='🎤 indie lyrics (200K docs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.METAL_LYRICS,
            display_modality=Modalities.TEXT,
            display_name='🎤 metal lyrics (200K docs)',
        ),
    ],
    Modalities.VIDEO: [
        DemoDataset(
            name=DemoDatasetNames.TUMBLR_GIFS,
            display_modality=Modalities.VIDEO,
            display_name='🎦 tumblr gifs (100K gifs)',
        ),
        DemoDataset(
            name=DemoDatasetNames.TUMBLR_GIFS_10K,
            display_modality=Modalities.VIDEO,
            display_name='🎦 tumblr gifs subset (10K gifs)',
        ),
    ],
}
DEFAULT_EXAMPLE_HOSTED = {
    'text_to_image': [
        DemoDatasetNames.BEST_ARTWORKS,
        DemoDatasetNames.DEEP_FASHION,
    ],
    'image_to_text': [DemoDatasetNames.RAP_LYRICS],
    'image_to_image': [DemoDatasetNames.TLL],
    'text_to_text': [DemoDatasetNames.ROCK_LYRICS],
}



import os.path
import pathlib
import tempfile
from multiprocessing import Process
from time import sleep
from typing import Dict

from jina import Flow
from jina.clients import Client
from kubernetes import client as k8s_client
from kubernetes import config
from yaspin.spinners import Spinners

from now.cloud_manager import is_local_cluster
from now.constants import DEFAULT_FLOW_NAME
from now.deployment.deployment import apply_replace, cmd, deploy_wolf
from now.log import time_profiler, yaspin_extended
from now.utils import sigmap, write_env_file, write_flow_file

cur_dir = pathlib.Path(__file__).parent.resolve()
MAX_WAIT_TIME = 1800


def batch(data_list, n=1):
    l = len(data_list)
    for ndx in range(0, l, n):
        yield data_list[ndx : min(ndx + n, l)]


def wait_for_lb(lb_name, ns):
    config.load_kube_config()
    v1 = k8s_client.CoreV1Api()
    while True:
        try:
            services = v1.list_namespaced_service(namespace=ns)
            ip = [
                s.status.load_balancer.ingress[0].ip
                for s in services.items
                if s.metadata.name == lb_name
            ][0]
            if ip:
                break
        except Exception:
            pass
        sleep(1)
    return ip


def check_pods_health(ns):
    config.load_kube_config()
    v1 = k8s_client.CoreV1Api()
    pods = v1.list_namespaced_pod(ns).items

    for pod in pods:
        try:
            message = pod.status.container_statuses[0].state.waiting.message
        except:
            message = None

        if message and 'Error' in message:
            raise Exception(pod.metadata.name + " " + message)


def deploy_k8s(f, ns, tmpdir, kubectl_path):
    k8_path = os.path.join(tmpdir, f'k8s/{ns}')
    with yaspin_extended(
        sigmap=sigmap, text="Convert Flow to Kubernetes YAML", color="green"
    ) as spinner:
        f.to_k8s_yaml(k8_path)
        spinner.ok('🔄')

    # create namespace
    cmd(f'{kubectl_path} create namespace {ns}')

    # deploy flow
    with yaspin_extended(
        Spinners.earth,
        sigmap=sigmap,
        text="Deploy Jina Flow (might take a bit)",
    ) as spinner:
        gateway_host_internal = f'gateway.{ns}.svc.cluster.local'
        gateway_port_internal = 8080
        if is_local_cluster(kubectl_path):
            apply_replace(
                f'{cur_dir}/k8s_backend-svc-node.yml',
                {'ns': ns},
                kubectl_path,
            )
            gateway_host = 'localhost'
            gateway_port = 31080
        else:
            apply_replace(f'{cur_dir}/k8s_backend-svc-lb.yml', {'ns': ns}, kubectl_path)
            gateway_host = wait_for_lb('gateway-lb', ns)
            gateway_port = 8080
        cmd(f'{kubectl_path} apply -R -f {k8_path}')

        client = Client(host=gateway_host, port=gateway_port)
        wait_time = 0
        while not client.is_flow_ready() and wait_time <= MAX_WAIT_TIME:
            check_pods_health(ns)
            wait_time += 1
            sleep(1)
        if not client.is_flow_ready():
            raise Exception('Flow execution timed out.')
        spinner.ok("🚀")

    return gateway_host, gateway_port, gateway_host_internal, gateway_port_internal


def start_flow_in_process(f):
    def start_flow():
        with f:
            print('flow started in process')
            f.block()

    p1 = Process(target=start_flow, args=())
    p1.daemon = False
    p1.start()


@time_profiler
def deploy_flow(
    deployment_type: str,
    flow_yaml: str,
    env_dict: Dict,
    kubectl_path: str,
):
    """Deploy a Flow on JCloud, Kubernetes, or using Jina Orchestration"""
    # TODO create tmpdir top level and pass it down
    with tempfile.TemporaryDirectory() as tmpdir:
        env_file = os.path.join(tmpdir, 'dot.env')
        write_env_file(env_file, env_dict)

        # hack we don't know if the flow yaml is a path or a string
        if type(flow_yaml) == dict:
            flow_file = os.path.join(tmpdir, 'flow.yml')
            write_flow_file(flow_yaml, flow_file)
            flow_yaml = flow_file

        if os.environ.get('NOW_TESTING', False):
            from dotenv import load_dotenv

            load_dotenv(env_file, override=True)

            f = Flow.load_config(flow_yaml)
            f.gateway_args.timeout_send = -1
            start_flow_in_process(f)

            host = 'localhost'
            client = Client(host=host, port=8080)

            # host & port
            gateway_host = 'remote'
            gateway_port = 8080
            gateway_host_internal = host
            gateway_port_internal = None  # Since host contains protocol

        elif deployment_type == 'remote':
            flow = deploy_wolf(path=flow_yaml)
            host = flow.endpoints['gateway']
            client = Client(host=host)

            # host & port
            gateway_host = 'remote'
            gateway_port = None
            gateway_host_internal = host
            gateway_port_internal = None  # Since host contains protocol
        else:
            from dotenv import load_dotenv

            load_dotenv(env_file, override=True)
            f = Flow.load_config(flow_yaml)
            (
                gateway_host,
                gateway_port,
                gateway_host_internal,
                gateway_port_internal,
            ) = deploy_k8s(
                f,
                DEFAULT_FLOW_NAME,
                tmpdir,
                kubectl_path=kubectl_path,
            )
            client = Client(host=gateway_host, port=gateway_port)

        if os.path.exists(env_file):
            os.remove(env_file)

    return (
        client,
        gateway_host,
        gateway_port,
        gateway_host_internal,
        gateway_port_internal,
    )



import json
import pathlib
import warnings

import cowsay
import docker
from kubernetes import client, config

from now.common.options import NEW_CLUSTER
from now.deployment.deployment import cmd
from now.log import time_profiler, yaspin_extended
from now.now_dataclasses import UserInput
from now.utils import maybe_prompt_user, sigmap

cur_dir = pathlib.Path(__file__).parent.resolve()
warnings.filterwarnings("ignore", category=DeprecationWarning)


def create_local_cluster(kind_path, **kwargs):
    out, err = cmd(f'{kind_path} get clusters')
    if err and 'No kind clusters' not in err.decode('utf-8'):
        print(err.decode('utf-8'))
        exit()
    cluster_name = 'jina-now'
    if cluster_name in out.decode('utf-8'):
        questions = [
            {
                'type': 'list',
                'name': 'proceed',
                'message': 'The local cluster is running already. '
                'Should it be recreated?',
                'choices': [
                    {'name': '⛔ no', 'value': False},
                    {'name': '✅ yes', 'value': True},
                ],
            },
        ]
        recreate = maybe_prompt_user(questions, 'proceed', **kwargs)
        if recreate:
            with yaspin_extended(
                sigmap=sigmap, text="Remove local cluster", color="green"
            ) as spinner:
                cmd(f'{kind_path} delete clusters {cluster_name}')
                spinner.ok('💀')
        else:
            cowsay.cow('see you soon 👋')
            exit(0)
    with yaspin_extended(
        sigmap=sigmap, text="Setting up local cluster", color="green"
    ) as spinner:
        kindest_images = docker.from_env().images.list('kindest/node')
        if len(kindest_images) == 0:
            print(
                'Download kind image to set up local cluster - this might take a while :)'
            )
        _, err = cmd(
            f'{kind_path} create cluster --name {cluster_name} --config {cur_dir}/kind.yml',
        )
        if err and 'failed to create cluster' in err.decode('utf-8'):
            print('\n' + err.decode('utf-8').split('ERROR')[-1])
            exit(1)
        spinner.ok("📦")


def is_local_cluster(kubectl_path):
    command = f'{kubectl_path} get nodes -o json'
    out, error = cmd(f'{kubectl_path} get nodes -o json')
    try:
        out = json.loads(out)
    except:
        print(f'Command {command} gives the following error: {error.decode("utf-8")}')
        exit(1)
    addresses = out['items'][0]['status']['addresses']
    is_local = len([a for a in addresses if a['type'] == 'ExternalIP']) == 0
    return is_local


@time_profiler
def setup_cluster(
    user_input: UserInput,
    kubectl_path='kubectl',
    kind_path='kind',
    **kwargs,
):
    if user_input.cluster == NEW_CLUSTER['value']:
        create_local_cluster(kind_path, **kwargs)
    elif user_input.deployment_type != 'remote':
        cmd(f'{kubectl_path} config use-context {user_input.cluster}')
        ask_existing(kubectl_path)


def ask_existing(kubectl_path, **kwargs):
    config.load_kube_config()
    v1 = client.CoreV1Api()
    if 'nowapi' in [item.metadata.name for item in v1.list_namespace().items]:
        with yaspin_extended(
            sigmap=sigmap, text="Remove old deployment", color="green"
        ) as spinner:
            cmd(f'{kubectl_path} delete ns nowapi')
            spinner.ok('💀')



from collections import defaultdict, namedtuple
from typing import Dict, List, Optional, Union

from docarray import Document, DocumentArray

SemanticScore = namedtuple(
    'SemanticScores',
    [
        'query_field',
        'document_field',
        'encoder',
        'linear_weight',
    ],
)
metrics_mapping = {
    'cosine': 'cosineSimilarity',
    'l2_norm': 'l2norm',
}


def generate_semantic_scores(
    docs_map: Dict[str, DocumentArray],
    encoder_to_fields: Dict[str, Union[List[str], str]],
) -> List[SemanticScore]:
    """Generate semantic scores from document mappings."""
    semantic_scores = []
    # either take fields names from _metadata, if multimodal doc
    # or take modality from tags of root doc
    for executor_name, da in docs_map.items():
        first_doc = da[0]
        if first_doc._metadata:  # must be a multimodal doc
            field_names = first_doc._metadata['multi_modal_schema'].keys()
        else:  # must be a unimodal doc
            field_names = [first_doc.tags['modality']]
        try:
            document_fields = encoder_to_fields[executor_name]
        except KeyError as e:
            raise KeyError(
                'Documents are not encoded with same encoder as query.'
            ) from e
        for field_name in field_names:
            for document_field in document_fields:
                semantic_scores.append(
                    SemanticScore(
                        query_field=field_name,
                        document_field=document_field,
                        encoder=executor_name,
                        linear_weight=1,
                    )
                )

    return semantic_scores


def build_es_queries(
    docs_map,
    apply_default_bm25: bool,
    get_score_breakdown: bool,
    semantic_scores: List[SemanticScore],
    custom_bm25_query: Optional[dict] = None,
    metric: Optional[str] = 'cosine',
    filter: dict = {},
    query_to_curated_ids: Dict[str, list] = {},
) -> Dict:
    """
    Build script-score query used in Elasticsearch. To do this, we extract
    embeddings from the query document and pass them in the script-score
    query together with the fields to search on in the Elasticsearch index.
    The query document will be returned with all of its embeddings as tags with
    their corresponding field+encoder as key.

    :param docs_map: dictionary mapping encoder to DocumentArray.
    :param apply_default_bm25: whether to combine bm25 with vector search. If False,
        will only perform vector search. If True, must supply a text
        field for bm25 searching.
    :param get_score_breakdown: whether to return the score breakdown for matches.
        For this function, this parameter determines whether to return the embeddings
        of a query document.
    :param custom_bm25_query: custom query to use for BM25.
    :param metric: metric to use for vector search.
    :param filter: dictionary of filters to apply to the search.
    :return: a dictionary containing query and filter.
    """
    queries = {}
    pinned_queries = {}
    docs = {}
    sources = {}
    script_params = defaultdict(dict)
    for executor_name, da in docs_map.items():
        for doc in da:
            if doc.id not in docs:
                docs[doc.id] = doc
                docs[doc.id].tags['embeddings'] = {}

            if doc.id not in queries:
                queries[doc.id] = get_default_query(
                    doc,
                    apply_default_bm25,
                    semantic_scores,
                    custom_bm25_query,
                    filter,
                )
                pinned_queries[doc.id] = get_pinned_query(
                    doc,
                    query_to_curated_ids,
                )

                if apply_default_bm25 or custom_bm25_query:
                    sources[doc.id] = '1.0 + _score / (_score + 10.0)'
                else:
                    sources[doc.id] = '1.0'

            for (
                query_field,
                document_field,
                encoder,
                linear_weight,
            ) in get_scores(executor_name, semantic_scores):
                field_doc = getattr(doc, query_field)
                if get_score_breakdown:
                    docs[doc.id].tags['embeddings'][
                        f'{query_field}-{encoder}'
                    ] = field_doc.embedding

                query_string = f'params.query_{query_field}_{executor_name}'
                document_string = f'{document_field}-{encoder}'

                sources[
                    doc.id
                ] += f" + {float(linear_weight)}*{metrics_mapping[metric]}({query_string}, '{document_string}.embedding')"

                script_params[doc.id][
                    f'query_{query_field}_{executor_name}'
                ] = field_doc.embedding

    es_queries = []

    for doc_id, query in queries.items():
        script_score = {
            'script_score': {
                'query': {
                    'bool': query['bool'],
                },
                'script': {
                    'source': sources[doc_id],
                    'params': script_params[doc_id],
                },
            },
        }
        if pinned_queries[doc_id]:
            query_json = {'pinned': pinned_queries[doc_id]['pinned']}
            query_json['pinned']['organic'] = script_score
        else:
            query_json = script_score
        es_queries.append((docs[doc_id], query_json))
    return es_queries


def get_default_query(
    doc: Document,
    apply_default_bm25: bool,
    semantic_scores: List[SemanticScore],
    custom_bm25_query: Dict = None,
    filter: Dict = {},
):
    query = {
        'bool': {
            'should': [
                {'match_all': {}},
            ],
        },
    }

    # build bm25 part
    if apply_default_bm25:
        bm25_semantic_score = next((x for x in semantic_scores if x.encoder == 'bm25'))
        if not bm25_semantic_score:
            raise ValueError(
                'No bm25 semantic scores found. Please specify this in the default_semantic_scores.'
            )
        text = getattr(doc, bm25_semantic_score.query_field).text
        multi_match = {'multi_match': {'query': text, 'fields': ['bm25_text']}}
        query['bool']['should'].append(multi_match)
    elif custom_bm25_query:
        query['bool']['should'].append(custom_bm25_query)

    # add filter
    if filter:
        es_search_filter = process_filter(filter)
        query['bool']['filter'] = es_search_filter

    return query


def get_pinned_query(doc: Document, query_to_curated_ids: Dict[str, list] = {}) -> Dict:
    pinned_query = {}
    if getattr(doc, 'query_text', None):
        query_text = doc.query_text.text
        if query_text in query_to_curated_ids.keys():
            pinned_query = {'pinned': {'ids': query_to_curated_ids[query_text]}}
    return pinned_query


def process_filter(filter) -> dict:
    es_search_filter = {}
    for field, filters in filter.items():
        for operator, filter in filters.items():
            field = field.replace('__', '.')
            if isinstance(filter, str):
                es_search_filter['term'] = {field: filter}
            elif isinstance(filter, int) or isinstance(filter, float):
                operator = operator.replace('$', '')
                es_search_filter['range'] = {field: {operator: filter}}
    return es_search_filter


def get_scores(encoder, semantic_scores):
    for (
        query_field,
        document_field,
        _encoder,
        linear_weight,
    ) in semantic_scores:
        if encoder == _encoder:
            yield query_field, document_field, _encoder, linear_weight



from typing import Dict

import numpy as np
from docarray import Document, DocumentArray
from docarray.array.chunk import ChunkArray


def merge_subdocuments(docs_map, encoder_to_fields) -> Dict[str, DocumentArray]:
    """
    Preprocess the docs_map to make it convertible to ES. Each field of a
    document should have only one subdocument. If the field document is a
    ChunkArray, only one embedding will be taken/created, and if the
    ChunkArray contains texts, these will be concatenated into one text.
    As for URIs, only the first one will be taken.

    :param docs_map: dictionary mapping encoder to DocumentArray.
    :param encoder_to_fields: dictionary mapping encoder to fields.
    :return: a dictionary mapping encoder to (processed) DocumentArray.
    """
    for executor_name, da in docs_map.items():
        field_names = encoder_to_fields[executor_name]
        for doc in da:
            for field_name in field_names:
                field_doc = getattr(doc, field_name)
                if isinstance(field_doc, ChunkArray):
                    # average the embeddings of the subdocuments
                    # TODO: use a better way to aggregate embeddings (eg. k-means)
                    embedding = average_embeddings_of_subdocuments(field_doc)
                    merged_doc = Document(
                        text=' '.join(
                            field_doc.texts
                        ).strip(),  # concatenate texts of subdocuments
                        embedding=embedding,
                    )
                    if field_doc[0].uri:
                        merged_doc.uri = field_doc[0].uri
                    merged_doc.tags['embeddings'] = {}
                    merged_doc.tags['embeddings'][
                        f'{field_name}-{executor_name}'
                    ] = field_doc[
                        ..., 'embedding'
                    ]  # will be a matrix, stacked
                    # replace ChunkArray with Document
                    setattr(doc, field_name, merged_doc)

    return docs_map


def average_embeddings_of_subdocuments(field_doc):
    return np.mean([chunk.embedding for chunk in field_doc], axis=0)



import json
import os
import pathlib
import time
from os.path import expanduser as user
from typing import Dict, List, Optional, Tuple

import hubble
from docarray import DocumentArray
from jina import __version__ as jina_version
from jina.helper import random_port

from now.app.base.app import JinaNOWApp
from now.constants import (
    EXECUTOR_PREFIX,
    EXTERNAL_CLIP_HOST,
    NOW_AUTOCOMPLETE_VERSION,
    NOW_ELASTIC_INDEXER_VERSION,
    NOW_PREPROCESSOR_VERSION,
    NOW_QDRANT_INDEXER_VERSION,
    PREFETCH_NR,
    TAG_INDEXER_DOC_HAS_TEXT,
    Modalities,
)
from now.demo_data import DEFAULT_EXAMPLE_HOSTED
from now.deployment.deployment import cmd
from now.executor.name_to_id_map import name_to_id_map
from now.now_dataclasses import UserInput

cur_dir = pathlib.Path(__file__).parent.resolve()


MAX_RETRIES = 20


def common_get_flow_env_dict(
    encoder_uses: str,
    encoder_with: Dict,
    encoder_uses_with: Dict,
    pre_trained_embedding_size: int,
    indexer_uses: str,
    indexer_resources: Dict,
    user_input: UserInput,
    tags: List,
    deployment_type: str,
):
    """Returns dictionary for the environments variables for the clip flow.yml files."""
    use_high_performance_flow = (
        get_email().split('@')[-1] == 'jina.ai' and deployment_type == 'remote'
    )
    if use_high_performance_flow:
        print(
            f"🚀🚀🚀 As employee of Jina, you are using a high performance flow.\n"
            f"Therefore, your deployment will be faster.\n"
            f"But make sure to scale it down after deployment."
        )
    config = {
        'JINA_VERSION': jina_version,
        'ENCODER_NAME': f'{EXECUTOR_PREFIX}{encoder_uses}',
        'CAST_CONVERT_NAME': f'{EXECUTOR_PREFIX}CastNMoveNowExecutor/v0.0.3',
        'N_DIM': pre_trained_embedding_size,
        'PRE_TRAINED_EMBEDDINGS_SIZE': pre_trained_embedding_size,
        'INDEXER_NAME': f'{EXECUTOR_PREFIX}{indexer_uses}',
        'PREFETCH': PREFETCH_NR,
        'PREPROCESSOR_NAME': f'{EXECUTOR_PREFIX}{name_to_id_map.get("NOWPreprocessor")}/{NOW_PREPROCESSOR_VERSION}',
        'PREPROCESSOR_REPLICAS': 15 if use_high_performance_flow else 1,
        'AUTOCOMPLETE_EXECUTOR_NAME': f'{EXECUTOR_PREFIX}{name_to_id_map.get("NOWAutoCompleteExecutor2")}/{NOW_AUTOCOMPLETE_VERSION}',
        'COLUMNS': tags,
        'ADMIN_EMAILS': user_input.admin_emails or [] if user_input.secured else [],
        'USER_EMAILS': user_input.user_emails or [] if user_input.secured else [],
        'API_KEY': [user_input.api_key]
        if user_input.secured and user_input.api_key
        else [],
        **encoder_with,
        **indexer_resources,
    }

    if encoder_uses_with.get('pretrained_model_name_or_path'):
        config['PRE_TRAINED_MODEL_NAME'] = encoder_uses_with[
            "pretrained_model_name_or_path"
        ]

    config['CUSTOM_DNS'] = ''
    if 'NOW_EXAMPLES' in os.environ:
        valid_app = DEFAULT_EXAMPLE_HOSTED.get(user_input.app_instance.app_name, {})
        is_demo_ds = user_input.dataset_name in valid_app
        if is_demo_ds:
            config[
                'CUSTOM_DNS'
            ] = f'now-example-{user_input.app_instance.app_name}-{user_input.dataset_name}.dev.jina.ai'
            config['CUSTOM_DNS'] = config['CUSTOM_DNS'].replace('_', '-')

    return config


def common_setup(
    app_instance: JinaNOWApp,
    user_input: UserInput,
    dataset: DocumentArray,
    encoder_uses: str,
    encoder_uses_with: Dict,
    indexer_uses: str,
    pre_trained_embedding_size: int,
    kubectl_path: str,
    encoder_with: Optional[Dict] = {},
    indexer_resources: Optional[Dict] = {},
) -> Dict:
    tags = _extract_tags_for_indexer(user_input)
    env_dict = common_get_flow_env_dict(
        encoder_uses=encoder_uses,
        encoder_with=encoder_with,
        encoder_uses_with=encoder_uses_with,
        pre_trained_embedding_size=pre_trained_embedding_size,
        indexer_uses=indexer_uses,
        indexer_resources=indexer_resources,
        user_input=user_input,
        tags=tags,
        deployment_type=user_input.deployment_type,
    )
    app_instance.set_flow_yaml(dataset_len=len(dataset))
    return env_dict


def get_email():
    try:
        with open(user('~/.jina/config.json')) as fp:
            config_val = json.load(fp)
            user_token = config_val['auth_token']
            client = hubble.Client(token=user_token, max_retries=None, jsonify=True)
            response = client.get_user_info()
        if 'email' in response['data']:
            return response['data']['email']
        return ''
    except FileNotFoundError:
        return ''


def get_indexer_config(
    elastic: Optional[bool] = False,
    kubectl_path: str = None,
    deployment_type: str = None,
) -> Dict:
    """Depending on the number of samples, which will be indexed, indexer and its resources are determined.

    :param num_indexed_samples: number of samples which will be indexed; should incl. chunks for e.g. text-to-video app
    :param elastic: hack to use NOWElasticIndexer, should be changed in future.
    :param kubectl_path: path to kubectl binary
    :param deployment_type: deployment type, e.g. 'remote' or 'local'
    :return: dict with indexer and its resource config
    """

    if elastic and deployment_type == 'local':
        config = {
            'indexer_uses': f'{name_to_id_map.get("NOWElasticIndexer")}/{NOW_ELASTIC_INDEXER_VERSION}',
            'hosts': setup_elastic_service(kubectl_path),
        }
    elif elastic and deployment_type == 'remote':
        raise ValueError(
            'NOWElasticIndexer is currently not supported for remote deployment. Please use local deployment.'
        )
    else:
        config = {
            'indexer_uses': f'{name_to_id_map.get("NOWQdrantIndexer16")}/{NOW_QDRANT_INDEXER_VERSION}'
        }

    config['indexer_resources'] = {'INDEXER_CPU': 0.5, 'INDEXER_MEM': '4G'}

    return config


def _extract_tags_for_indexer(user_input: UserInput):
    final_tags = []
    for tag, value in user_input.filter_fields_modalities.items():
        if tag in user_input.filter_fields:
            final_tags.append([tag, value])
    if user_input.app_instance.output_modality in [
        Modalities.IMAGE,
        Modalities.VIDEO,
    ]:
        final_tags.append([TAG_INDEXER_DOC_HAS_TEXT, str(bool.__name__)])
    return final_tags


def setup_elastic_service(
    kubectl_path: str,
) -> str:
    """Setup ElasticSearch service and return a connection string to connect to the service with.

    :param kubectl_path: path to kubectl binary
    :return: connection string for connecting to the ElasticSearch service.
    """
    cur_dir = pathlib.Path(__file__).parent.resolve()
    cmd(
        f'{kubectl_path} create -f https://download.elastic.co/downloads/eck/2.4.0/crds.yaml'
    )
    cmd(
        f'{kubectl_path} apply -f https://download.elastic.co/downloads/eck/2.4.0/operator.yaml'
    )
    cmd(f'{kubectl_path} create ns nowapi')
    cmd(f'{kubectl_path} apply -f {cur_dir}/../deployment/elastic_kind.yml')
    num_retries = 0
    es_password, error_msg = '', b''
    while num_retries < MAX_RETRIES:
        es_password, error_msg = cmd(
            [
                kubectl_path,
                "get",
                "secret",
                "quickstart-es-elastic-user",
                "-o",
                "go-template='{{.data.elastic | base64decode}}'",
            ]
        )
        if es_password:
            es_password = es_password.decode("utf-8")[1:-1]
            break
        else:
            num_retries += 1
            time.sleep(2)
    if not es_password:
        raise Exception(error_msg.decode("utf-8"))
    host = f"https://elastic:{es_password}@quickstart-es-http.default:9200"
    return host


def _get_clip_apps_with_dict(user_input: UserInput) -> Tuple[Dict, Dict]:
    """Depending on whether this app will be remotely deployed, this function returns the with
    dictionary for the CLIP executor."""
    is_remote = user_input.deployment_type == 'remote'
    encoder_with = {
        'ENCODER_HOST': EXTERNAL_CLIP_HOST if is_remote else '0.0.0.0',
        'ENCODER_PORT': 443 if is_remote else random_port(),
        'IS_REMOTE_DEPLOYMENT': is_remote,
    }
    return encoder_with



import json
import re

import requests
from jina.serve.runtimes.gateway.http.models import JinaResponseModel
from pydantic import BaseModel, parse_obj_as

from deployment.bff.app.v1.models.search import (
    IndexRequestModel,
    SearchRequestModel,
    SearchResponseModel,
)
from deployment.bff.app.v1.routers.helper import field_dict_to_doc, jina_client_post
from now.common.options import construct_app


class Client:
    """
    This is the jina NOW client which can be used to run requests against apps which are deployed using jcloud.
    """

    def __init__(self, jcloud_id, app, api_key):
        self.jcloud_id = jcloud_id
        self.app = app
        self.api_key = api_key

    def send_request_bff(self, endpoint: str, **kwargs):
        request_body = {
            "host": f'grpcs://nowapi-{self.jcloud_id}.wolf.jina.ai',
            "api_key": self.api_key,
            **kwargs,
        }
        response = requests.post(
            f'https://nowrun.jina.ai/api/v1/{endpoint}',
            json=request_body,
        )
        return response

    def send_request(self, endpoint: str, **kwargs):
        """
        Client to run requests against a deployed flow
        """
        if endpoint != 'search':
            raise NotImplementedError('Only search endpoint is supported for now')

        if 'text' in kwargs:
            query_doc = field_dict_to_doc({'text_field': {'text': kwargs.pop('text')}})

        app_request = SearchRequestModel(
            host=f'grpcs://nowapi-{self.jcloud_id}.wolf.jina.ai',
            api_key=self.api_key,
            **kwargs,
        )
        response = jina_client_post(
            app_request,
            endpoint,
            inputs=query_doc,
            parameters={'limit': app_request.limit, 'filter': app_request.filters},
        )
        return response



import os
import uuid
from collections import defaultdict
from copy import deepcopy

from docarray import Document, DocumentArray

from now.app.base.app import JinaNOWApp
from now.constants import DatasetTypes
from now.data_loading.elasticsearch import ElasticsearchExtractor
from now.log import yaspin_extended
from now.now_dataclasses import UserInput
from now.utils import sigmap


def load_data(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    """Based on the user input, this function will pull the configured DocumentArray dataset ready for the preprocessing
    executor.

    :param app: chosen JinaNOWApp
    :param user_input: The configured user object. Result from the Jina Now cli dialog.
    :return: The loaded DocumentArray.
    """
    da = None
    if user_input.dataset_type == DatasetTypes.DOCARRAY:
        print('⬇  Pull DocumentArray dataset')
        da = _pull_docarray(user_input.dataset_name)
    elif user_input.dataset_type == DatasetTypes.PATH:
        print('💿  Loading files from disk')
        da = _load_from_disk(app, user_input)
        da = _load_tags_from_json_if_needed(da, user_input)
    elif user_input.dataset_type == DatasetTypes.S3_BUCKET:
        da = _list_files_from_s3_bucket(app=app, user_input=user_input)
        da = _load_tags_from_json_if_needed(da, user_input)
    elif user_input.dataset_type == DatasetTypes.ELASTICSEARCH:
        da = _extract_es_data(user_input)
    elif user_input.dataset_type == DatasetTypes.DEMO:
        print('⬇  Download DocumentArray dataset')
        da = DocumentArray.pull(name=user_input.dataset_name, show_progress=True)
    if da is None:
        raise ValueError(
            f'Could not load DocumentArray dataset. Please check your configuration: {user_input}.'
        )
    if 'NOW_CI_RUN' in os.environ:
        da = da[:50]
    return da


def select_ending(files, endings):
    for file in files:
        for ending in endings:
            if file.endswith(ending):
                return file
    return None


def _load_tags_from_json_if_needed(da: DocumentArray, user_input: UserInput):
    if any([doc.uri.endswith('.json') for doc in da]):
        return _load_tags_from_json(da, user_input)
    else:
        return da


def _load_tags_from_json(da, user_input):
    print(
        f'Loading tags! We assume that you have a folder for each document. The folder contains a content '
        f'file (image, text, video, ...) and a json file containing the tags'
    )
    # map folders to all files they contain
    folder_to_files = defaultdict(list)
    for d in da:
        folder = d.uri.rsplit('/', 1)[0]
        folder_to_files[folder].append(d.uri)

    docs = DocumentArray()
    for files in folder_to_files.values():
        tag_file = select_ending(files, ['json'])
        content_file = select_ending(
            files, user_input.app_instance.supported_file_types
        )
        if content_file:
            if tag_file:
                tags = {'tag_uri': tag_file}
            else:
                tags = {}
            docs.append(Document(uri=content_file, tags=tags))
    return docs


def _pull_docarray(dataset_name: str):
    try:
        return DocumentArray.pull(name=dataset_name, show_progress=True)
    except Exception:
        raise ValueError(
            '💔 oh no, the secret of your docarray is wrong, or it was deleted after 14 days'
        )


def _load_to_datauri_and_save_into_tags(d: Document) -> Document:
    d.tags['uri'] = d.uri
    return d.convert_uri_to_datauri()


def match_types(uri, supported_file_types):
    for t in supported_file_types:
        if t == '**' or uri.split('.')[-1] == t:
            return True
    return False


def _extract_es_data(user_input: UserInput) -> DocumentArray:
    query = {
        'query': {'match_all': {}},
        '_source': True,
    }
    es_extractor = ElasticsearchExtractor(
        query=query,
        index=user_input.es_index_name,
        connection_str=user_input.es_host_name,
    )
    extracted_docs = es_extractor.extract(search_fields=user_input.search_fields)
    return extracted_docs


def _load_from_disk(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    dataset_path = user_input.dataset_path.strip()
    dataset_path = os.path.expanduser(dataset_path)
    if os.path.isfile(dataset_path):
        try:
            return DocumentArray.load_binary(dataset_path)
        except Exception as e:
            print(f'Failed to load the binary file provided under path {dataset_path}')
            exit(1)
    elif os.path.isdir(dataset_path):
        with yaspin_extended(
            sigmap=sigmap, text="Loading data from folder", color="green"
        ) as spinner:
            spinner.ok('🏭')
            docs = DocumentArray.from_files(f'{dataset_path}/**')
            docs = DocumentArray(
                d
                for d in docs
                if match_types(d.uri, app.supported_file_types + ['json'])
            )
            docs.apply(_load_to_datauri_and_save_into_tags)
            return docs
    else:
        raise ValueError(
            f'The provided dataset path {dataset_path} does not'
            f' appear to be a valid file or folder on your system.'
        )


def _list_files_from_s3_bucket(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    bucket, folder_prefix = get_s3_bucket_and_folder_prefix(user_input)
    docs = []
    with yaspin_extended(
        sigmap=sigmap, text="Listing files from S3 bucket ...", color="green"
    ) as spinner:
        spinner.ok('🏭')
        for obj in list(bucket.objects.filter(Prefix=folder_prefix)):
            file_type = str(obj.key).split('.')
            if file_type[-1] in app.supported_file_types + ['json']:
                docs.append(Document(uri=f"s3://{bucket.name}/{obj.key}"))
    return DocumentArray(docs)


def get_s3_bucket_and_folder_prefix(user_input: UserInput):
    import boto3.session

    s3_uri = user_input.dataset_path
    if not s3_uri.startswith('s3://'):
        raise ValueError(
            f"Can't process S3 URI {s3_uri} as it assumes it starts with: 's3://'"
        )

    bucket = s3_uri.split('/')[2]
    folder_prefix = '/'.join(s3_uri.split('/')[3:])

    session = boto3.session.Session(
        aws_access_key_id=user_input.aws_access_key_id,
        aws_secret_access_key=user_input.aws_secret_access_key,
    )
    bucket = session.resource('s3').Bucket(bucket)

    return bucket, folder_prefix


def deep_copy_da(da: DocumentArray) -> DocumentArray:
    new_da = DocumentArray()
    for i, d in enumerate(da):
        new_doc = deepcopy(d)
        new_doc.id = str(uuid.uuid4())
        new_da.append(new_doc)
    return new_da



import base64
from typing import List

from docarray import Document, DocumentArray
from fastapi import APIRouter

from deployment.bff.app.v1.models.search import (
    IndexRequestModel,
    SearchRequestModel,
    SearchResponseModel,
    SuggestionRequestModel,
)
from deployment.bff.app.v1.routers.helper import field_dict_to_doc, jina_client_post

router = APIRouter()


@router.post(
    "/index",
    summary='Add more data to the indexer',
)
def index(data: IndexRequestModel):
    index_docs = DocumentArray()
    for field_dict, tags_dict in data.data:
        doc = field_dict_to_doc(field_dict)
        doc.tags.update(tags_dict)
        index_docs.append(doc)

    jina_client_post(
        data=data,
        inputs=index_docs,
        endpoint='/index',
    )


@router.post(
    "/search",
    response_model=List[SearchResponseModel],
    summary='Search data via query',
)
def search(data: SearchRequestModel):
    query_doc = field_dict_to_doc(data.query)

    query_filter = {key: {'$eq': value} for key, value in data.filters.items()}

    docs = jina_client_post(
        endpoint='/search',
        inputs=query_doc,
        parameters={'limit': data.limit, 'filter': query_filter},
        data=data,
    )
    matches = []
    for doc in docs[0].matches:
        # todo: use multimodal doc in the future
        scores = {}
        for score_name, named_score in doc.scores.items():
            scores[score_name] = named_score.to_dict()
        if doc.uri:
            result = {'uri': doc.uri}
        elif doc.blob:
            result = {'blob': base64.b64encode(doc.blob).decode('utf-8')}
        elif doc.text:
            result = {'text': doc.text}
        match = SearchResponseModel(
            id=doc.id,
            scores=scores,
            tags=doc.tags,
            fields={'result_field': result},
        )
        matches.append(match)
    return matches


@router.post(
    "/suggestion",
    summary='Get auto complete suggestion for query',
)
def suggestion(data: SuggestionRequestModel):
    suggest_doc = Document(text=data.text)
    docs = jina_client_post(
        endpoint='/suggestion',
        inputs=suggest_doc,
        data=data,
    )
    return docs.to_dict()



from jina import Client

flow_id = '...'
api_key = ''
jwt = ''

client = Client(host=f'grpcs://nowapi-{flow_id}.wolf.jina.ai')

client.post(
    '/curate',
    parameters={
        'api_key': api_key,
        'jwt': jwt,
        'query_to_filter': {
            'query1': [{'uri': {'$eq': '3'}}],
            'query2': [
                {'uri': {'$eq': 'uri2'}},
                {'tags__color': {'$eq': 'red'}},
            ],
        },
    },
)



import argparse

from jina.parsers.helper import _ColoredHelpFormatter

from now import __version__
from now.common import options
from now.common.options import construct_app
from now.constants import Apps


def set_base_parser():
    """Set the base parser
    :return: the parser
    """

    # create the top-level parser
    urls = {
        'Code': ('💻', 'https://github.com/jina-ai/now'),
        'Jina Docs': ('📖', 'https://docs.jina.ai'),
        'Help': ('💬', 'https://slack.jina.ai'),
        'Hiring!': ('🙌', 'https://career.jina.ai'),
    }
    url_str = '\n'.join(f'- {v[0]:<10} {k:10.10}\t{v[1]}' for k, v in urls.items())

    parser = argparse.ArgumentParser(
        epilog=f'Jina NOW - get your neural search case up in minutes. \n\n{url_str}',
        formatter_class=_chf,
        description='Command Line Interface of `%(prog)s`',
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=__version__,
        help='Show Jina version',
    )
    return parser


def set_help_parser(parser=None):
    """Set the parser for the jina help lookup
    :param parser: an optional existing parser to build upon
    :return: the parser
    """

    if not parser:
        from jina.parsers.base import set_base_parser

        parser = set_base_parser()

    parser.add_argument(
        'query',
        type=str,
        help='Lookup the usage & mention of the argument name in Jina NOW',
    )
    return parser


def set_start_parser(sp):
    """Add the arguments for the jina now to the parser
    :param parser: an optional existing parser to build upon
    :return: the parser
    """

    parser = sp.add_parser(
        'start',
        help='Start jina now and create or reuse a cluster.',
        description='Start jina now and create or reuse a cluster.',
        formatter_class=_chf,
    )

    # Get list of enabled apps
    enabled_apps, enabled_apps_instance = [], []
    for app in Apps():
        app_instance = construct_app(app)
        if app_instance.is_enabled:
            enabled_apps.append(app)
            enabled_apps_instance.append(app_instance)

    parser.add_argument(
        '--app',
        help='Select the app you would like to use. Do not use this argument when'
        ' using the `%(prog)-8s [sub-command]`',
        choices=enabled_apps,
        type=str,
    )

    # Add common app options
    for option in options.base_options:
        if option.is_terminal_command:
            _kwargs = {
                'help': option.description,
                'type': str,
            }
            _kwargs.update(option.argparse_kwargs)
            parser.add_argument(
                f'--{option.name}',
                **_kwargs,
            )

    # Add app sub-command and its options
    sub_parser = parser.add_subparsers(
        dest='app',
        description='use `%(prog)-8s [sub-command] --help` '
        'to get additional arguments to be used with each sub-command',
        required=False,
    )

    # Set parser args for the enabled apps
    for app_instance in enabled_apps_instance:
        app_instance.set_app_parser(sub_parser, formatter=_chf)


def set_stop_parser(sp):
    sp.add_parser(
        'stop',
        help='Stop jina now and remove local cluster.',
        description='Stop jina now and remove local cluster.',
        formatter_class=_chf,
    )


def set_survey_parser(sp):
    sp.add_parser(
        'survey',
        help='Opens a survey in the browser. Thanks for providing us feedback.',
        description='Opens a survey in the browser. Thanks for providing us feedback.',
        formatter_class=_chf,
    )


def get_main_parser():
    """The main parser for Jina NOW
    :return: the parser
    """
    # create the top-level parser
    parser = set_base_parser()
    sp = parser.add_subparsers(
        dest='cli',
        description='',
        required=True,
    )

    set_start_parser(sp)
    set_stop_parser(sp)
    set_survey_parser(sp)

    return parser


_chf = _ColoredHelpFormatter



import abc
import os
from typing import Dict, List, Optional, Tuple

import docker
from docarray import DocumentArray
from jina import Client
from jina.jaml import JAML

from now.app.base.preprocess import preprocess_image, preprocess_text, preprocess_video
from now.constants import DEFAULT_FLOW_NAME, SUPPORTED_FILE_TYPES, Modalities
from now.demo_data import AVAILABLE_DATASETS, DEFAULT_EXAMPLE_HOSTED, DemoDataset
from now.now_dataclasses import DialogOptions, UserInput


class JinaNOWApp:
    """
    Interface for Jina NOW applications.
    To create a new application, you need to inherit from this class and override the methods.
    Example:
        class TextToImageApp(JinaNowApp):
            pass
    """

    def __init__(self):
        self.flow_yaml = ''

    @property
    def app_name(self) -> str:
        """
        Name of the app. Should be an enum value set in now.constants.Apps
        """
        raise NotImplementedError()

    @property
    def is_enabled(self) -> bool:
        """
        Set to True if this app is enabled for the end user.
        """
        raise NotImplementedError()

    @property
    def description(self) -> str:
        """
        Short description of the app.
        """
        return 'Jina NOW app'

    @property
    @abc.abstractmethod
    def input_modality(self) -> List[Modalities]:
        """
        Modality used for running search queries
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def output_modality(self) -> List[Modalities]:
        """
        Modality used for indexing data
        """
        raise NotImplementedError()

    def set_flow_yaml(self, **kwargs):
        """Used to configure the flow yaml in the Jina NOW app.
        The interface is as follows:
        - if kwargs['finetuning']=True, choose finetuning flow
        - if kwargs['encode']=True, choose encoding flow (to get embeddings for finetuning)
        - temporarily introduced kwargs['dataset_len'], if app optimizes different flows to it
        """
        flow_dir = os.path.abspath(os.path.join(__file__, '..'))
        self.flow_yaml = os.path.join(flow_dir, 'flow.yml')

    @property
    def bff(self) -> Optional[str]:
        """
        TODO This function is currently not used but already introduces the concept of custom bff
        Used to configure the bff which is used to define input and output format.
        :return: the path to the file where the bff routs are configured
        """
        return None

    @property
    def playground(self) -> Optional[str]:
        """
        TODO This function is currently not used but already introduces the concept of custom playground
        Used to configure the playground(streamlit app) where the user can run example queries
        :return: the path to the streamlit file.
        """
        return None

    @property
    def options(self) -> List[DialogOptions]:
        """
        Get the options which are used to configure the app. Base class should override this function and
        return the list of their option. Check ``DialogOptions`` for the format of the options
        On CLI the user will get a prompt and at the storefront, a GUI will be generated accordingly.
        Example:
        ``return [options.your_custom_options`]``
        :return: List[DialogOptions]
        """
        return []

    @property
    def supported_file_types(self) -> List[str]:
        """Used to filter files in local structure or an S3 bucket."""
        sup_file = [SUPPORTED_FILE_TYPES[modality] for modality in self.output_modality]
        return [item for sublist in sup_file for item in sublist]

    @property
    def demo_datasets(self) -> Dict[str, List[DemoDataset]]:
        """Get a list of example datasets for the app."""
        available_datasets = {}
        for output_modality in self.output_modality:
            available_datasets[output_modality] = AVAILABLE_DATASETS[output_modality]
        return available_datasets

    @property
    def required_docker_memory_in_gb(self) -> int:
        """
        Recommended memory limit for the docker client to run this app.
        """
        return 8

    @property
    def finetune_datasets(self) -> [Tuple]:
        """Defines the list of demo datasets which are fine-tunable."""
        return ()

    def _check_docker_mem_limit(self) -> bool:
        mem_total = docker.from_env().info().get('MemTotal')
        if (
            mem_total is not None
            and mem_total / 1e9 < self.required_docker_memory_in_gb
        ):
            print(
                '🚨 Your docker container memory limit is set to ~{:.2f}GB'.format(
                    mem_total / 1e9
                )
                + f' which is below the recommended limit of {self.required_docker_memory_in_gb}GB'
                f' for the {self.app_name} app'
            )
            return False
        else:
            return True

    def set_app_parser(self, parser, formatter) -> None:
        """
        This parser reads from the `options` property and parses it
        to form the command line arguments for app
        """
        if self.is_enabled:
            parser = parser.add_parser(
                self.app_name,
                help=self.description,
                description=f'Create an {self.app_name} app.',
                formatter_class=formatter,
            )
            for option in self.options:
                if option.is_terminal_command:
                    parser.add_argument(
                        f'--{option.name}',
                        help=option.description,
                        type=str,
                    )

    def _check_requirements(self) -> bool:
        """
        Returns true if all requirements on the system are satisfied. Else False.
        """
        return True

    def run_checks(self, user_input: UserInput) -> bool:
        req_check = self._check_requirements()
        mem_check = True
        if user_input.deployment_type != 'remote':
            mem_check = self._check_docker_mem_limit()
        return req_check and mem_check

    # TODO Remove kubectl_path. At the moment, the setup function needs kubectl because of finetuning a custom
    #  dataset with local deployment. In that case, inference is done on the k8s cluster.
    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path: str
    ) -> Dict:
        """
        Runs before the flow is deployed.
        Common use cases:
            - create a database
            - finetune a model + push the artifact
            - notify other services
            - check if starting the app is currently possible
        :param dataset:
        :param user_input: user configuration based on the given options
        :return: dict used to replace variables in flow yaml and to clean up resources after the flow is terminated
        """
        with open(self.flow_yaml) as input_f:
            flow_yaml_content = JAML.load(input_f.read())
            flow_yaml_content['jcloud']['labels'] = {'team': 'now'}
            flow_yaml_content['jcloud']['name'] = (
                user_input.flow_name + '-' + DEFAULT_FLOW_NAME
                if user_input.flow_name != ''
                and user_input.flow_name != DEFAULT_FLOW_NAME
                else DEFAULT_FLOW_NAME
            )

            # append api_keys to the executor with name 'preprocessor' and 'indexer'
            for executor in flow_yaml_content['executors']:
                if executor['name'] == 'preprocessor' or executor['name'] == 'indexer':
                    executor['uses_with']['api_keys'] = '${{ ENV.API_KEY }}'

            if Modalities.TEXT in self.input_modality:
                if not any(
                    exec_dict['name'] == 'autocomplete_executor'
                    for exec_dict in flow_yaml_content['executors']
                ):
                    flow_yaml_content['executors'].insert(
                        0,
                        {
                            'name': 'autocomplete_executor',
                            'uses': '${{ ENV.AUTOCOMPLETE_EXECUTOR_NAME }}',
                            'needs': 'gateway',
                            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
                            'uses_with': {
                                'api_keys': '${{ ENV.API_KEY }}',
                                'user_emails': '${{ ENV.USER_EMAILS }}',
                                'admin_emails': '${{ ENV.ADMIN_EMAILS }}',
                            },
                        },
                    )
            self.add_environment_variables(flow_yaml_content)
            self.flow_yaml = flow_yaml_content
        return {}

    def preprocess(
        self,
        docs: DocumentArray,
    ) -> DocumentArray:
        """Loads and preprocesses every document such that it is ready for indexing."""
        for doc in docs:
            for chunk in doc.chunks:
                try:
                    if chunk.modality == 'text':
                        preprocess_text(chunk)
                    elif chunk.modality == 'image':
                        preprocess_image(chunk)
                    elif chunk.modality == 'video':
                        preprocess_video(chunk)
                    else:
                        raise ValueError(f'Unsupported modality {chunk.modality}')
                except Exception as e:
                    print(e)
        return docs

    def is_demo_available(self, user_input) -> bool:
        hosted_ds = DEFAULT_EXAMPLE_HOSTED.get(self.app_name, {})
        if (
            hosted_ds
            and user_input.dataset_name in hosted_ds
            and user_input.deployment_type == 'remote'
            and 'NOW_EXAMPLES' not in os.environ
            and 'NOW_CI_RUN' not in os.environ
        ):
            client = Client(
                host=f'grpcs://now-example-{self.app_name}-{user_input.dataset_name}.dev.jina.ai'.replace(
                    '_', '-'
                )
            )
            try:
                client.post('/dry_run', timeout=2)
            except Exception:
                return False
            return True
        return False

    @property
    def max_request_size(self) -> int:
        """Max number of documents in one request"""
        return 32

    def add_environment_variables(self, flow_yaml_content):
        if 'JINA_OPTOUT_TELEMETRY' in os.environ:
            flow_yaml_content['with']['env']['JINA_OPTOUT_TELEMETRY'] = os.environ[
                'JINA_OPTOUT_TELEMETRY'
            ]
            for executor in flow_yaml_content['executors']:
                executor['env']['JINA_OPTOUT_TELEMETRY'] = os.environ[
                    'JINA_OPTOUT_TELEMETRY'
                ]



from concurrent.futures import ProcessPoolExecutor
from time import time

from jina import Client, Document

flow_id = ''
api_key = ''
search_text = ''
limit = 60


client = Client(host=f'grpcs://nowapi-{flow_id}.wolf.jina.ai')


def call():
    t0 = time()
    result = client.post(
        '/search',
        inputs=Document(chunks=Document(text=search_text)),
        parameters={
            'api_key': api_key,
            'limit': limit,
        },
    )
    return time() - t0


# measure latency
start = time()
for i in range(10):
    call()
print(f'Latency : {(time() - start) / 10}s')


# measure QPS
num_queries = 100
worker = 5
with ProcessPoolExecutor(max_workers=worker) as executor:
    # QPS test
    start = time()
    futures = []
    latencies = []

    for i in range(num_queries):
        future = executor.submit(call)
        futures.append(future)
    for future in futures:
        latencies.append(future.result())
    print(f'QPS: {num_queries / (time() - start)}s')

latencies = sorted(latencies)
for p in [0, 50, 75, 85, 90, 95, 99, 99.9]:
    print(f"P{p}: {latencies[int(len(latencies) * p / 100)]}")
